#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import concurrent.futures
import difflib
import os
import sys
import traceback
from collections import defaultdict
from collections import OrderedDict
from datetime import datetime
from datetime import timedelta
from distutils.util import strtobool
from itertools import groupby
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Dict
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type

import humanize
from bravado.exception import BravadoConnectionError
from bravado.exception import BravadoTimeoutError
from bravado.exception import HTTPError
from service_configuration_lib import read_deploy

from paasta_tools import kubernetes_tools
from paasta_tools.adhoc_tools import AdhocJobConfig
from paasta_tools.api.client import get_paasta_api_client
from paasta_tools.cli.utils import execute_paasta_serviceinit_on_remote_master
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import get_instance_configs_for_service
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import NoSuchService
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.flink_tools import FlinkDeploymentConfig
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import KubernetesDeployStatus
from paasta_tools.marathon_serviceinit import bouncing_status_human
from paasta_tools.marathon_serviceinit import desired_state_human
from paasta_tools.marathon_serviceinit import haproxy_backend_report
from paasta_tools.marathon_serviceinit import marathon_app_deploy_status_human
from paasta_tools.marathon_serviceinit import status_marathon_job_human
from paasta_tools.marathon_tools import MarathonDeployStatus
from paasta_tools.mesos_tools import format_tail_lines_for_mesos_task
from paasta_tools.monitoring_tools import get_team
from paasta_tools.monitoring_tools import list_teams
from paasta_tools.tron_tools import TronActionConfig
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import datetime_from_utc_to_local
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import format_table
from paasta_tools.utils import get_soa_cluster_deploy_files
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import list_all_instances_for_service
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import load_deployments_json
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import remove_ansi_escape_sequences
from paasta_tools.utils import SystemPaastaConfig


HTTP_ONLY_INSTANCE_CONFIG: Sequence[Type[InstanceConfig]] = [
    FlinkDeploymentConfig,
    KubernetesDeploymentConfig,
    AdhocJobConfig,
]
SSH_ONLY_INSTANCE_CONFIG: Sequence[Type[InstanceConfig]] = []


def add_subparser(subparsers,) -> None:
    status_parser = subparsers.add_parser(
        "status",
        help="Display the status of a PaaSTA service.",
        description=(
            "'paasta status' works by SSH'ing to remote PaaSTA masters and "
            "inspecting the local APIs, and reports on the overal health "
            "of a service."
        ),
        epilog=(
            "Note: This command requires SSH and sudo privileges on the remote PaaSTA "
            "masters."
        ),
    )
    status_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help="Print out more output regarding the state of the service. "
        "A second -v will also print the stdout/stderr tail.",
    )
    status_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    add_instance_filter_arguments(status_parser)
    status_parser.set_defaults(command=paasta_status)


def add_instance_filter_arguments(status_parser, verb: str = "inspect") -> None:
    status_parser.add_argument(
        "-s", "--service", help=f"The name of the service you wish to {verb}"
    ).completer = lazy_choices_completer(list_services)
    status_parser.add_argument(
        "-c",
        "--clusters",
        help=f"A comma-separated list of clusters to {verb}. By default, will {verb} all clusters.\n"
        f"For example: --clusters norcal-prod,nova-prod",
    ).completer = lazy_choices_completer(list_clusters)
    status_parser.add_argument(
        "-i",
        "--instances",
        help=f"A comma-separated list of instances to {verb}. By default, will {verb} all instances.\n"
        f"For example: --instances canary,main",
    )  # No completer because we need to know service first and we can't until some other stuff has happened
    status_parser.add_argument(
        "-l",
        "--deploy-group",
        help=(
            f"Name of the deploy group which you want to {verb}. "
            f"If specified together with --instances and/or --clusters, will {verb} common instances only."
        ),
    ).completer = lazy_choices_completer(list_deploy_groups)
    status_parser.add_argument(
        "-o",
        "--owner",
        help=f"Only {verb} instances with this owner specified in soa-configs.",
    ).completer = lazy_choices_completer(list_teams)
    status_parser.add_argument(
        "-r", "--registration", help=f"Only {verb} instances with this registration."
    )


def missing_deployments_message(service: str,) -> str:
    message = (
        f"{service} has no deployments in deployments.json yet.\n  " "Has Jenkins run?"
    )
    return message


def get_deploy_info(deploy_file_path: str,) -> Mapping:
    deploy_info = read_deploy(deploy_file_path)
    if not deploy_info:
        paasta_print("Error encountered with %s" % deploy_file_path)

        exit(1)
    return deploy_info


def get_planned_deployments(service: str, soa_dir: str) -> Iterable[str]:
    for cluster, cluster_deploy_file in get_soa_cluster_deploy_files(
        service=service, soa_dir=soa_dir
    ):
        for instance in get_deploy_info(cluster_deploy_file):
            yield f"{cluster}.{instance}"


def list_deployed_clusters(
    pipeline: Sequence[str], actual_deployments: Sequence[str]
) -> Sequence[str]:
    """Returns a list of clusters that a service is deployed to given
    an input deploy pipeline and the actual deployments"""
    deployed_clusters: List[str] = []
    for namespace in pipeline:
        cluster, instance = namespace.split(".")
        if namespace in actual_deployments:
            if cluster not in deployed_clusters:
                deployed_clusters.append(cluster)
    return deployed_clusters


def get_actual_deployments(service: str, soa_dir: str) -> Mapping[str, str]:
    deployments_json = load_deployments_json(service, soa_dir)
    if not deployments_json:
        paasta_print(
            "Warning: it looks like %s has not been deployed anywhere yet!" % service,
            file=sys.stderr,
        )
    # Create a dictionary of actual $service Jenkins deployments
    actual_deployments = {}
    for key, branch_dict in deployments_json.config_dict.items():
        service, namespace = key.split(":")
        if service == service:
            value = branch_dict["docker_image"]
            sha = value[value.rfind("-") + 1 :]
            actual_deployments[namespace.replace("paasta-", "", 1)] = sha
    return actual_deployments


def paasta_status_on_api_endpoint(
    cluster: str,
    service: str,
    instance: str,
    output: List[str],
    system_paasta_config: SystemPaastaConfig,
    verbose: int,
) -> int:
    output.append("    instance: %s" % PaastaColors.blue(instance))
    client = get_paasta_api_client(cluster, system_paasta_config)
    if not client:
        paasta_print("Cannot get a paasta-api client")
        exit(1)
    try:
        status = client.service.status_instance(
            service=service, instance=instance, verbose=verbose
        ).result()
    except HTTPError as exc:
        output.append(PaastaColors.red(exc.response.text))
        return exc.status_code
    except (BravadoConnectionError, BravadoTimeoutError) as exc:
        output.append(
            PaastaColors.red(f"Could not connect to API: {exc.__class__.__name__}")
        )
        return 1
    except Exception:
        tb = sys.exc_info()[2]
        output.append(PaastaColors.red(f"Exception when talking to the API:"))
        output.extend(line.strip() for line in traceback.format_tb(tb))
        return 1

    if status.git_sha != "":
        output.append("    Git sha:    %s (desired)" % status.git_sha)

    if status.marathon is not None:
        return print_marathon_status(service, instance, output, status.marathon)
    elif status.kubernetes is not None:
        return print_kubernetes_status(service, instance, output, status.kubernetes)
    elif status.tron is not None:
        return print_tron_status(service, instance, output, status.tron, verbose)
    elif status.adhoc is not None:
        return print_adhoc_status(
            cluster, service, instance, output, status.adhoc, verbose
        )
    elif status.flink is not None:
        return print_flink_status(
            cluster, service, instance, output, status.flink, verbose
        )
    else:
        paasta_print(
            "Not implemented: Looks like %s is not a Marathon or Kubernetes instance"
            % instance
        )
        return 0


def print_adhoc_status(
    cluster: str,
    service: str,
    instance: str,
    output: List[str],
    status,
    verbose: int = 0,
) -> int:
    output.append(f"    Job: {instance}")
    for run in status:
        output.append(
            "Launch time: %s, run id: %s, framework id: %s"
            % (run["launch_time"], run["run_id"], run["framework_id"])
        )
    if status:
        output.append(
            (
                "    Use `paasta remote-run stop -s {} -c {} -i {} [-R <run id> "
                "    | -F <framework id>]` to stop."
            ).format(service, cluster, instance)
        )
    else:
        output.append("    Nothing found.")

    return 0


def print_marathon_status(
    service: str, instance: str, output: List[str], marathon_status
) -> int:
    if marathon_status.error_message:
        output.append(marathon_status.error_message)
        return 1

    bouncing_status = bouncing_status_human(
        marathon_status.app_count, marathon_status.bounce_method
    )
    desired_state = desired_state_human(
        marathon_status.desired_state, marathon_status.expected_instance_count
    )
    output.append(f"    Desired state:      {bouncing_status} and {desired_state}")

    job_status_human = status_marathon_job_human(
        service=service,
        instance=instance,
        deploy_status=marathon_status.deploy_status,
        desired_app_id=marathon_status.desired_app_id,
        app_count=marathon_status.app_count,
        running_instances=marathon_status.running_instance_count,
        normal_instance_count=marathon_status.expected_instance_count,
    )
    output.append(f"    {job_status_human}")

    if marathon_status.autoscaling_info:
        autoscaling_info_table = create_autoscaling_info_table(
            marathon_status.autoscaling_info
        )
        output.extend([f"      {line}" for line in autoscaling_info_table])

    for app_status in marathon_status.app_statuses:
        app_status_human = marathon_app_status_human(
            marathon_status.desired_app_id, app_status
        )
        output.extend([f"      {line}" for line in app_status_human])

    mesos_status_human = marathon_mesos_status_human(
        marathon_status.mesos.error_message,
        marathon_status.mesos.running_task_count or 0,
        marathon_status.expected_instance_count,
        marathon_status.mesos.running_tasks,
        marathon_status.mesos.non_running_tasks,
    )
    output.extend([f"    {line}" for line in mesos_status_human])

    if marathon_status.smartstack is not None:
        smartstack_status_human = marathon_smartstack_status_human(
            marathon_status.smartstack.registration,
            marathon_status.smartstack.expected_backends_per_location,
            marathon_status.smartstack.locations,
            # TODO: this is just to avoid rollout issues where the client updates
            # before the API server.  This can be removed after it first rolls
            # out
            getattr(marathon_status.smartstack, "error_message", None),
        )
        output.extend([f"    {line}" for line in smartstack_status_human])

    return 0


autoscaling_fields_to_headers = OrderedDict(
    current_instances="Current instances",
    max_instances="Max instances",
    min_instances="Min instances",
    current_utilization="Current utilization",
    target_instances="Target instances",
)


def create_autoscaling_info_table(autoscaling_info):
    output = ["Autoscaling Info:"]

    if autoscaling_info.current_utilization is not None:
        autoscaling_info.current_utilization = "{:.1f}%".format(
            autoscaling_info.current_utilization * 100
        )
    else:
        autoscaling_info.current_utilization = "Exception"

    if autoscaling_info.target_instances is None:
        autoscaling_info.target_instances = "Exception"

    headers = list(autoscaling_fields_to_headers.values())
    row = [
        str(getattr(autoscaling_info, field)) for field in autoscaling_fields_to_headers
    ]
    table = [f"  {line}" for line in format_table([headers, row])]
    output.extend(table)
    return output


def marathon_mesos_status_human(
    error_message,
    running_task_count,
    expected_instance_count,
    running_tasks,
    non_running_tasks,
):
    if error_message:
        return [f"Mesos: {PaastaColors.red(error_message)}"]

    output = []
    output.append(
        marathon_mesos_status_summary(running_task_count, expected_instance_count)
    )

    if running_tasks or non_running_tasks:
        output.append("  Running Tasks:")
        running_tasks_table = create_mesos_running_tasks_table(running_tasks)
        output.extend([f"    {line}" for line in running_tasks_table])

        output.append(PaastaColors.grey("  Non-running Tasks:"))
        non_running_tasks_table = create_mesos_non_running_tasks_table(
            non_running_tasks
        )
        output.extend([f"    {line}" for line in non_running_tasks_table])

    return output


def create_mesos_running_tasks_table(running_tasks):
    rows = []
    table_header = [
        "Mesos Task ID",
        "Host deployed to",
        "Ram",
        "CPU",
        "Deployed at what localtime",
    ]
    rows.append(table_header)
    for task in running_tasks or []:
        mem_string = get_mesos_task_memory_string(task)
        cpu_string = get_mesos_task_cpu_string(task)
        deployed_at = datetime.fromtimestamp(task.deployed_timestamp)
        deployed_at_string = "{} ({})".format(
            deployed_at.strftime("%Y-%m-%dT%H:%M"), humanize.naturaltime(deployed_at)
        )

        rows.append(
            [task.id, task.hostname, mem_string, cpu_string, deployed_at_string]
        )
        rows.extend(format_tail_lines_for_mesos_task(task.tail_lines, task.id))

    return format_table(rows)


def get_mesos_task_memory_string(task):
    if task.rss.value is None or task.mem_limit.value is None:
        return task.rss.error_message or task.mem_limit.error_message
    elif task.mem_limit.value == 0:
        return "Undef"
    else:
        mem_percent = 100 * task.rss.value / task.mem_limit.value
        mem_string = "%d/%dMB" % (
            (task.rss.value / 1024 / 1024),
            (task.mem_limit.value / 1024 / 1024),
        )
        if mem_percent > 90:
            return PaastaColors.red(mem_string)
        else:
            return mem_string


def get_mesos_task_cpu_string(task):
    if task.cpu_shares.value is None or task.cpu_used_seconds.value is None:
        return task.cpu_shares.error_message
    else:
        # The total time a task has been allocated is the total time the task has
        # been running multiplied by the "shares" a task has.
        # (see https://github.com/mesosphere/mesos/blob/0b092b1b0/src/webui/master/static/js/controllers.js#L140)
        allocated_seconds = task.cpu_shares.value * task.duration_seconds
        if allocated_seconds == 0:
            return "Undef"
        else:
            cpu_percent = round(
                100 * (task.cpu_used_seconds.value / allocated_seconds), 1
            )
            cpu_string = "%s%%" % cpu_percent
            if cpu_percent > 90:
                return PaastaColors.red(cpu_string)
            else:
                return cpu_string


def create_mesos_non_running_tasks_table(non_running_tasks):
    rows = []
    table_header = [
        "Mesos Task ID",
        "Host deployed to",
        "Deployed at what localtime",
        "Status",
    ]
    rows.append(table_header)

    for task in non_running_tasks or []:
        if task.deployed_timestamp is None:
            deployed_at_string = "Unknown"
        else:
            deployed_at = datetime.fromtimestamp(task.deployed_timestamp)
            deployed_at_string = "{} ({})".format(
                deployed_at.strftime("%Y-%m-%dT%H:%M"),
                humanize.naturaltime(deployed_at),
            )

        rows.append([task.id, task.hostname, deployed_at_string, task.state])
        rows.extend(format_tail_lines_for_mesos_task(task.tail_lines, task.id))

    table = format_table(rows)
    return [PaastaColors.grey(formatted_row) for formatted_row in table]


def marathon_mesos_status_summary(mesos_task_count, expected_instance_count) -> str:
    if mesos_task_count >= expected_instance_count:
        status = PaastaColors.green("Healthy")
        count_str = PaastaColors.green(
            "(%d/%d)" % (mesos_task_count, expected_instance_count)
        )
    elif mesos_task_count == 0:
        status = PaastaColors.red("Critical")
        count_str = PaastaColors.red(
            "(%d/%d)" % (mesos_task_count, expected_instance_count)
        )
    else:
        status = PaastaColors.yellow("Warning")
        count_str = PaastaColors.yellow(
            "(%d/%d)" % (mesos_task_count, expected_instance_count)
        )
    running_string = PaastaColors.bold("TASK_RUNNING")
    return f"Mesos:      {status} - {count_str} tasks in the {running_string} state."


def marathon_app_status_human(app_id, app_status) -> List[str]:
    output = []

    if app_status.dashboard_url:
        output.append(f"Dashboard: {PaastaColors.blue(app_status.dashboard_url)}")
    else:
        output.append(f"App ID: {PaastaColors.blue(app_id)}")

    output.append(
        "  "
        + " ".join(
            [
                f"{app_status.tasks_running} running,",
                f"{app_status.tasks_healthy} healthy,",
                f"{app_status.tasks_staged} staged",
                f"out of {app_status.tasks_total}",
            ]
        )
    )

    create_datetime = datetime.fromtimestamp(app_status.create_timestamp)
    output.append(
        "  App created: {} ({})".format(
            create_datetime, humanize.naturaltime(create_datetime)
        )
    )

    deploy_status = MarathonDeployStatus.fromstring(app_status.deploy_status)
    deploy_status_human = marathon_app_deploy_status_human(
        deploy_status, app_status.backoff_seconds
    )
    output.append(f"  Status: {deploy_status_human}")

    if app_status.tasks:
        output.append("  Tasks:")
        tasks_table = format_marathon_task_table(app_status.tasks)
        output.extend([f"    {line}" for line in tasks_table])

    if app_status.unused_offer_reason_counts is not None:
        output.append("  Possibly stalled for:")
        output.extend(
            [
                f"    {reason}: {count}"
                for reason, count in app_status.unused_offer_reason_counts.items()
            ]
        )

    return output


def format_marathon_task_table(tasks):
    rows = [
        ("Mesos Task ID", "Host deployed to", "Deployed at what localtime", "Health")
    ]
    for task in tasks:
        local_deployed_datetime = datetime_from_utc_to_local(
            datetime.fromtimestamp(task.deployed_timestamp)
        )
        if task.host is not None:
            hostname = f"{task.host}:{task.port}"
        else:
            hostname = "Unknown"

        if task.is_healthy is None:
            health_check_status = PaastaColors.grey("N/A")
        elif task.is_healthy:
            health_check_status = PaastaColors.green("Healthy")
        else:
            health_check_status = PaastaColors.red("Unhealthy")

        rows.append(
            (
                task.id,
                hostname,
                "{} ({})".format(
                    local_deployed_datetime.strftime("%Y-%m-%dT%H:%M"),
                    humanize.naturaltime(local_deployed_datetime),
                ),
                health_check_status,
            )
        )

    return format_table(rows)


def format_kubernetes_pod_table(pods):
    rows = [("Pod ID", "Host deployed to", "Deployed at what localtime", "Health")]
    for pod in pods:
        local_deployed_datetime = datetime_from_utc_to_local(
            datetime.fromtimestamp(pod.deployed_timestamp)
        )
        hostname = f"{pod.host}" if pod.host is not None else "Unknown"

        if pod.phase is None or pod.phase == "Pending":
            health_check_status = PaastaColors.grey("N/A")
        elif pod.phase == "Running":
            health_check_status = PaastaColors.green("Healthy")
        else:
            health_check_status = PaastaColors.red("Unhealthy")

        rows.append(
            (
                pod.name,
                hostname,
                "{} ({})".format(
                    local_deployed_datetime.strftime("%Y-%m-%dT%H:%M"),
                    humanize.naturaltime(local_deployed_datetime),
                ),
                health_check_status,
            )
        )

    return format_table(rows)


def marathon_smartstack_status_human(
    registration, expected_backends_per_location, locations, error_message
) -> List[str]:
    if error_message:
        return [f"Smartstack: {PaastaColors.red(error_message)}"]
    elif len(locations) == 0:
        return [f"Smartstack: ERROR - {registration} is NOT in smartstack at all!"]

    output = ["Smartstack:"]
    output.append(f"  Haproxy Service Name: {registration}")
    output.append(f"  Backends:")
    for location in locations:
        backend_status = haproxy_backend_report(
            expected_backends_per_location, location.running_backends_count
        )
        output.append(f"    {location.name} - {backend_status}")

        if location.backends:
            backends_table = build_smartstack_backends_table(location.backends)
            output.extend([f"      {line}" for line in backends_table])

    return output


def build_smartstack_backends_table(backends):
    rows = [("Name", "LastCheck", "LastChange", "Status")]
    for backend in backends:
        if backend.status == "UP":
            status = PaastaColors.default(backend.status)
        elif backend.status == "DOWN":
            status = PaastaColors.red(backend.status)
        elif backend.status == "MAINT":
            status = PaastaColors.grey(backend.status)
        else:
            status = PaastaColors.yellow(backend.status)

        if backend.check_duration is None:
            check_duration = ""
        else:
            check_duration = str(backend.check_duration)

        row = (
            f"{backend.hostname}:{backend.port}",
            f"{backend.check_status}/{backend.check_code} in {check_duration}ms",
            humanize.naturaltime(timedelta(seconds=backend.last_change)),
            status,
        )

        if not backend.has_associated_task:
            row = tuple(
                PaastaColors.grey(remove_ansi_escape_sequences(col)) for col in row
            )

        rows.append(row)

    return format_table(rows)


def kubernetes_app_deploy_status_human(status, backoff_seconds=None):
    status_string = kubernetes_tools.KubernetesDeployStatus.tostring(status)

    if status == kubernetes_tools.KubernetesDeployStatus.Waiting:
        deploy_status = (
            "%s (new tasks waiting for capacity to become available)"
            % PaastaColors.red(status_string)
        )
    elif status == kubernetes_tools.KubernetesDeployStatus.Deploying:
        deploy_status = PaastaColors.yellow(status_string)
    elif status == kubernetes_tools.KubernetesDeployStatus.Running:
        deploy_status = PaastaColors.bold(status_string)
    else:
        deploy_status = status_string

    return deploy_status


def status_kubernetes_job_human(
    service: str,
    instance: str,
    deploy_status: str,
    desired_app_id: str,
    app_count: int,
    running_instances: int,
    normal_instance_count: int,
) -> str:
    name = PaastaColors.cyan(compose_job_id(service, instance))

    if app_count >= 0:
        if running_instances >= normal_instance_count:
            status = PaastaColors.green("Healthy")
            instance_count = PaastaColors.green(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        elif running_instances == 0:
            status = PaastaColors.yellow("Critical")
            instance_count = PaastaColors.red(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        else:
            status = PaastaColors.yellow("Warning")
            instance_count = PaastaColors.yellow(
                "(%d/%d)" % (running_instances, normal_instance_count)
            )
        return "Kubernetes:   {} - up with {} instances. Status: {}".format(
            status, instance_count, deploy_status
        )
    else:
        status = PaastaColors.yellow("Warning")
        return "Kubernetes:   {} - {} (app {}) is not configured in Kubernetes yet (waiting for bounce)".format(
            status, name, desired_app_id
        )


def print_flink_status(
    cluster: str,
    service: str,
    instance: str,
    output: List[str],
    flink: Mapping[str, Any],
    verbose: int,
) -> int:
    status = flink.get("status")
    if status is None:
        output.append(PaastaColors.red("    Flink cluster is not available yet"))
        return 1

    # Since metadata should be available no matter the state, we show it first. If this errors out
    # then we cannot really do much to recover, because cluster is not in usable state anyway
    metadata = flink.get("metadata")
    config_sha = metadata.labels.get("yelp.com/paasta_config_sha")
    if config_sha is None:
        raise ValueError(f"expected config sha on Flink, but received {metadata}")
    if config_sha.startswith("config"):
        config_sha = config_sha[6:]

    output.append(f"    Config SHA: {config_sha}")

    if status.state != "running":
        output.append(
            "    State: {state}".format(state=PaastaColors.yellow(status.state))
        )
        output.append(f"    No other information available in non-running state")
        return 0

    dashboard_url = metadata.annotations.get("yelp.com/dashboard_url")
    if verbose:
        output.append(
            f"    Flink version: {status.config['flink-version']} {status.config['flink-revision']}"
        )
    else:
        output.append(f"    Flink version: {status.config['flink-version']}")
    output.append(f"    URL: {dashboard_url}/")
    output.append(f"    State: {status.state}")
    output.append(
        "    Jobs:"
        f" {status.overview['jobs-running']} running,"
        f" {status.overview['jobs-finished']} finished,"
        f" {status.overview['jobs-failed']} failed,"
        f" {status.overview['jobs-cancelled']} cancelled"
    )
    output.append(
        "   "
        f" {status.overview['taskmanagers']} taskmanagers,"
        f" {status.overview['slots-available']}/{status.overview['slots-total']} slots available"
    )

    output.append(f"    Jobs:")
    if verbose:
        output.append(
            f"      Job Name                         State       Job ID                           Started"
        )
    else:
        output.append(f"      Job Name                         State       Started")

    # Use only the most recent jobs
    unique_jobs = (
        sorted(jobs, key=lambda j: -j["start-time"])[0]
        for _, jobs in groupby(
            sorted(
                (j for j in status.jobs if j.get("name") and j.get("start-time")),
                key=lambda j: j["name"],
            ),
            lambda j: j["name"],
        )
    )
    for job in unique_jobs:
        job_id = job["jid"]
        if verbose:
            fmt = """      {job_name: <32.32} {state: <11} {job_id} {start_time}
        {dashboard_url}"""
        else:
            fmt = "      {job_name: <32.32} {state: <11} {start_time}"
        start_time = datetime_from_utc_to_local(
            datetime.utcfromtimestamp(int(job["start-time"]) // 1000)
        )
        output.append(
            fmt.format(
                job_id=job_id,
                job_name=job["name"].split(".", 2)[2],
                state=(job.get("state") or "unknown"),
                start_time=f"{str(start_time)} ({humanize.naturaltime(start_time)})",
                dashboard_url=PaastaColors.grey(f"{dashboard_url}/#/jobs/{job_id}"),
            )
        )
        if verbose and job_id in status.exceptions:
            exceptions = status.exceptions[job_id]
            root_exception = exceptions["root-exception"]
            if root_exception is not None:
                output.append(f"        Exception: {root_exception}")
                ts = exceptions["timestamp"]
                if ts is not None:
                    exc_ts = datetime_from_utc_to_local(
                        datetime.utcfromtimestamp(int(ts) // 1000)
                    )
                    output.append(
                        f"            {str(exc_ts)} ({humanize.naturaltime(exc_ts)})"
                    )
    return 0


def print_kubernetes_status(
    service: str, instance: str, output: List[str], kubernetes_status
) -> int:
    if kubernetes_status.error_message:
        output.append(kubernetes_status.error_message)
        return 1

    bouncing_status = bouncing_status_human(
        kubernetes_status.app_count, kubernetes_status.bounce_method
    )
    desired_state = desired_state_human(
        kubernetes_status.desired_state, kubernetes_status.expected_instance_count
    )
    output.append(f"    State:      {bouncing_status} - Desired state: {desired_state}")

    status = KubernetesDeployStatus.fromstring(kubernetes_status.deploy_status)
    deploy_status = kubernetes_app_deploy_status_human(status)

    output.append(
        "    {}".format(
            status_kubernetes_job_human(
                service=service,
                instance=instance,
                deploy_status=deploy_status,
                desired_app_id=kubernetes_status.app_id,
                app_count=kubernetes_status.app_count,
                running_instances=kubernetes_status.running_instance_count,
                normal_instance_count=kubernetes_status.expected_instance_count,
            )
        )
    )
    if kubernetes_status.create_timestamp:
        create_datetime = datetime.fromtimestamp(kubernetes_status.create_timestamp)
        output.append(
            "      App created: {} ({}). Namespace: {}".format(
                create_datetime,
                humanize.naturaltime(create_datetime),
                kubernetes_status.namespace,
            )
        )

    if kubernetes_status.pods and len(kubernetes_status.pods) > 0:
        output.append("      Pods:")
        pods_table = format_kubernetes_pod_table(kubernetes_status.pods)
        output.extend([f"        {line}" for line in pods_table])
    return 0


def print_tron_status(
    service: str, instance: str, output: List[str], tron_status, verbose: int = 0
) -> int:
    output.append(f"    Tron job: {tron_status.job_name}")
    if verbose:
        output.append(f"      Status: {tron_status.job_status}")
        output.append(f"      Schedule: {tron_status.job_schedule}")
    output.append("      Dashboard: {}".format(PaastaColors.blue(tron_status.job_url)))

    output.append(f"    Action: {tron_status.action_name}")
    output.append(f"      Status: {tron_status.action_state}")
    if verbose:
        output.append(f"      Start time: {tron_status.action_start_time}")
    output.append(f"      Command: {tron_status.action_command}")
    if verbose > 1:
        output.append(f"      Raw Command: {tron_status.action_raw_command}")
        output.append(f"      Stdout: \n{tron_status.action_stdout}")
        output.append(f"      Stderr: \n{tron_status.action_stderr}")

    return 0


def report_status_for_cluster(
    service: str,
    cluster: str,
    deploy_pipeline: Sequence[str],
    actual_deployments: Mapping[str, str],
    instance_whitelist: Mapping[str, Type[InstanceConfig]],
    system_paasta_config: SystemPaastaConfig,
    verbose: int = 0,
    use_api_endpoint: bool = False,
) -> Tuple[int, Sequence[str]]:
    """With a given service and cluster, prints the status of the instances
    in that cluster"""
    output = ["", "service: %s" % service, "cluster: %s" % cluster]
    seen_instances = []
    deployed_instances = []
    instances = instance_whitelist.keys()
    http_only_instances = [
        instance
        for instance, instance_config_class in instance_whitelist.items()
        if instance_config_class in HTTP_ONLY_INSTANCE_CONFIG
    ]
    ssh_only_instances = [
        instance
        for instance, instance_config_class in instance_whitelist.items()
        if instance_config_class in SSH_ONLY_INSTANCE_CONFIG
    ]

    tron_jobs = [
        instance
        for instance, instance_config_class in instance_whitelist.items()
        if instance_config_class == TronActionConfig
    ]

    for namespace in deploy_pipeline:
        cluster_in_pipeline, instance = namespace.split(".")
        seen_instances.append(instance)

        if cluster_in_pipeline != cluster:
            continue
        if instances and instance not in instances:
            continue

        # Case: service deployed to cluster.instance
        if namespace in actual_deployments:
            deployed_instances.append(instance)

        # Case: flink instances don't use `deployments.json`
        elif instance_whitelist.get(instance) == FlinkDeploymentConfig:
            deployed_instances.append(instance)

        # Case: service NOT deployed to cluster.instance
        else:
            output.append("  instance: %s" % PaastaColors.red(instance))
            output.append("    Git sha:    None (not deployed yet)")

    api_return_code = 0
    ssh_return_code = 0
    if len(deployed_instances) > 0:
        http_only_deployed_instances = [
            deployed_instance
            for deployed_instance in deployed_instances
            if (
                deployed_instance in http_only_instances
                or deployed_instance not in ssh_only_instances
                and use_api_endpoint
            )
        ]
        if len(http_only_deployed_instances):
            return_codes = [
                paasta_status_on_api_endpoint(
                    cluster=cluster,
                    service=service,
                    instance=deployed_instance,
                    output=output,
                    system_paasta_config=system_paasta_config,
                    verbose=verbose,
                )
                for deployed_instance in http_only_deployed_instances
            ]
            if any(return_codes):
                api_return_code = 1
        ssh_only_deployed_instances = [
            deployed_instance
            for deployed_instance in deployed_instances
            if (
                deployed_instance in ssh_only_instances
                or deployed_instance not in http_only_instances
                and not use_api_endpoint
            )
        ]
        if len(ssh_only_deployed_instances):
            ssh_return_code, status = execute_paasta_serviceinit_on_remote_master(
                "status",
                cluster,
                service,
                ",".join(
                    deployed_instance
                    for deployed_instance in ssh_only_deployed_instances
                ),
                system_paasta_config,
                stream=False,
                verbose=verbose,
                ignore_ssh_output=True,
            )
            # Status results are streamed. This print is for possible error messages.
            if status is not None:
                for line in status.rstrip().split("\n"):
                    output.append("    %s" % line)

    if len(tron_jobs) > 0:
        return_codes = [
            paasta_status_on_api_endpoint(
                cluster=cluster,
                service=service,
                instance=tron_job,
                output=output,
                system_paasta_config=system_paasta_config,
                verbose=verbose,
            )
            for tron_job in tron_jobs
        ]
        seen_instances.extend(tron_jobs)

    output.append(
        report_invalid_whitelist_values(instances, seen_instances, "instance")
    )

    if ssh_return_code:
        return_code = ssh_return_code
    elif api_return_code:
        return_code = api_return_code
    else:
        return_code = 0

    return return_code, output


def report_invalid_whitelist_values(
    whitelist: Iterable[str], items: Sequence[str], item_type: str
) -> str:
    """Warns the user if there are entries in ``whitelist`` which don't
    correspond to any item in ``items``. Helps highlight typos.
    """
    return_string = ""
    bogus_entries = []
    if whitelist is None:
        return ""
    for entry in whitelist:
        if entry not in items:
            bogus_entries.append(entry)
    if len(bogus_entries) > 0:
        return_string = (
            "\n" "Warning: This service does not have any %s matching these names:\n%s"
        ) % (item_type, ",".join(bogus_entries))
    return return_string


def verify_instances(
    args_instances: str, service: str, clusters: Sequence[str]
) -> Sequence[str]:
    """Verify that a list of instances specified by user is correct for this service.

    :param args_instances: a list of instances.
    :param service: the service name
    :param cluster: a list of clusters
    :returns: a list of instances specified in args_instances without any exclusions.
    """
    unverified_instances = args_instances.split(",")
    service_instances: Set[str] = list_all_instances_for_service(
        service, clusters=clusters
    )

    misspelled_instances: Sequence[str] = [
        i for i in unverified_instances if i not in service_instances
    ]

    if misspelled_instances:
        suggestions: List[str] = []
        for instance in misspelled_instances:
            matches = difflib.get_close_matches(
                instance, service_instances, n=5, cutoff=0.5
            )
            suggestions.extend(matches)  # type: ignore
        suggestions = list(set(suggestions))

        if clusters:
            message = "{} doesn't have any instances matching {} on {}.".format(
                service,
                ", ".join(sorted(misspelled_instances)),
                ", ".join(sorted(clusters)),
            )
        else:
            message = "{} doesn't have any instances matching {}.".format(
                service, ", ".join(sorted(misspelled_instances))
            )

        paasta_print(PaastaColors.red(message))

        if suggestions:
            paasta_print("Did you mean any of these?")
            for instance in sorted(suggestions):
                paasta_print("  %s" % instance)

    return unverified_instances


def normalize_registrations(
    service: str, registrations: Sequence[str]
) -> Sequence[str]:
    ret = []
    for reg in registrations:
        if "." not in reg:
            ret.append(f"{service}.{reg}")
        else:
            ret.append(reg)
    return ret


def get_filters(args,) -> Sequence[Callable[[InstanceConfig], bool]]:
    """Figures out which filters to apply from an args object, and returns them

    :param args: args object
    :returns: list of functions that take an instance config and returns if the instance conf matches the filter
    """
    filters = []

    if args.service:
        filters.append(lambda conf: conf.get_service() in args.service.split(","))

    if args.clusters:
        filters.append(lambda conf: conf.get_cluster() in args.clusters.split(","))

    if args.instances:
        filters.append(lambda conf: conf.get_instance() in args.instances.split(","))

    if args.deploy_group:
        filters.append(
            lambda conf: conf.get_deploy_group() in args.deploy_group.split(",")
        )

    if args.registration:
        normalized_regs = normalize_registrations(
            service=args.service, registrations=args.registration.split(",")
        )
        filters.append(
            lambda conf: any(
                reg in normalized_regs
                for reg in (
                    conf.get_registrations()
                    if hasattr(conf, "get_registrations")
                    else []
                )
            )
        )

    if args.owner:
        owners = args.owner.split(",")

        filters.append(
            # If the instance owner is None, check the service owner, else check the instance owner
            lambda conf: get_team(
                overrides={}, service=conf.get_service(), soa_dir=args.soa_dir
            )
            in owners
            if conf.get_team() is None
            else conf.get_team() in owners
        )

    return filters


def apply_args_filters(
    args,
) -> Mapping[str, Mapping[str, Mapping[str, Type[InstanceConfig]]]]:
    """
    Take an args object and returns the dict of cluster:service:instances
    Currently, will filter by clusters, instances, services, and deploy_groups
    If no instances are found, will print a message and try to find matching instances
    for each service

    :param args: args object containing attributes to filter by
    :returns: Dict of dicts, in format {cluster_name: {service_name: {instance1, instance2}}}
    """
    clusters_services_instances: DefaultDict[
        str, DefaultDict[str, Dict[str, Type[InstanceConfig]]]
    ] = defaultdict(lambda: defaultdict(dict))

    if args.service:
        try:
            validate_service_name(args.service, soa_dir=args.soa_dir)
        except NoSuchService:
            paasta_print(
                PaastaColors.red(f'The service "{args.service}" does not exist.')
            )
            all_services = list_services(soa_dir=args.soa_dir)
            suggestions = difflib.get_close_matches(
                args.service, all_services, n=5, cutoff=0.5
            )
            if suggestions:
                paasta_print(PaastaColors.red(f"Did you mean any of these?"))
                for suggestion in suggestions:
                    paasta_print(PaastaColors.red(f"  {suggestion}"))
            return clusters_services_instances

        all_services = [args.service]
    else:
        all_services = list_services(soa_dir=args.soa_dir)

    if args.service is None and args.owner is None:
        args.service = figure_out_service_name(args, soa_dir=args.soa_dir)

    if args.clusters:
        clusters = args.clusters.split(",")
    else:
        clusters = list_clusters()

    if args.instances:
        instances = args.instances.split(",")
    else:
        instances = None

    filters = get_filters(args)

    i_count = 0
    for service in all_services:
        if args.service and service != args.service:
            continue

        for instance_conf in get_instance_configs_for_service(
            service, soa_dir=args.soa_dir, clusters=clusters, instances=instances
        ):
            if all([f(instance_conf) for f in filters]):
                cluster_service = clusters_services_instances[
                    instance_conf.get_cluster()
                ][service]
                cluster_service[instance_conf.get_instance()] = instance_conf.__class__
                i_count += 1

    if i_count == 0 and args.service and args.instances:
        for service in args.service.split(","):
            verify_instances(args.instances, service, clusters)

    return clusters_services_instances


def paasta_status(args,) -> int:
    """Print the status of a Yelp service running on PaaSTA.
    :param args: argparse.Namespace obj created from sys.args by cli"""
    soa_dir = args.soa_dir
    system_paasta_config = load_system_paasta_config()

    if "USE_API_ENDPOINT" in os.environ:
        # bool will throw a ValueError if it doesn't recognize $USE_API_ENDPOINT
        use_api_endpoint = bool(strtobool(os.environ["USE_API_ENDPOINT"]))
    else:
        use_api_endpoint = False

    return_codes = [0]
    tasks = []
    clusters_services_instances = apply_args_filters(args)
    for cluster, service_instances in clusters_services_instances.items():
        for service, instances in service_instances.items():
            all_flink = all(i == FlinkDeploymentConfig for i in instances.values())
            actual_deployments: Mapping[str, str]
            if all_flink:
                actual_deployments = {}
            else:
                actual_deployments = get_actual_deployments(service, soa_dir)
            if all_flink or actual_deployments:
                deploy_pipeline = list(get_planned_deployments(service, soa_dir))
                tasks.append(
                    (
                        report_status_for_cluster,
                        dict(
                            service=service,
                            cluster=cluster,
                            deploy_pipeline=deploy_pipeline,
                            actual_deployments=actual_deployments,
                            instance_whitelist=instances,
                            system_paasta_config=system_paasta_config,
                            verbose=args.verbose,
                            use_api_endpoint=use_api_endpoint,
                        ),
                    )
                )
            else:
                paasta_print(missing_deployments_message(service))
                return_codes.append(1)

    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        tasks = [executor.submit(t[0], **t[1]) for t in tasks]  # type: ignore
        for future in concurrent.futures.as_completed(tasks):  # type: ignore
            return_code, output = future.result()
            paasta_print("\n".join(output))
            return_codes.append(return_code)

    return max(return_codes)



import logging
import time
from collections import namedtuple
from queue import Empty
from queue import PriorityQueue
from queue import Queue
from threading import Condition
from threading import Event
from threading import Thread
from typing import Any
from typing import Collection
from typing import List
from typing import Optional
from typing import Tuple

from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.marathon_tools import get_all_marathon_apps
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import load_marathon_service_config_no_cache
from paasta_tools.marathon_tools import MarathonClients
from paasta_tools.marathon_tools import MarathonServiceConfig
from paasta_tools.utils import load_system_paasta_config

BounceTimers = namedtuple(
    "BounceTimers", ["processed_by_worker", "setup_marathon", "bounce_length"]
)
BaseServiceInstance = namedtuple(
    "ServiceInstance",
    [
        "service",
        "instance",
        "bounce_by",
        "wait_until",
        "watcher",
        "bounce_timers",
        "failures",
        "processed_count",
    ],
)


class ServiceInstance(BaseServiceInstance):
    __slots__ = ()

    def __new__(
        _cls,
        service: str,
        instance: str,
        watcher: str,
        cluster: str,
        bounce_by: float,
        wait_until: float,
        failures: int = 0,
        bounce_timers: Optional[BounceTimers] = None,
        processed_count: int = 0,
    ) -> "ServiceInstance":
        return super().__new__(  # type: ignore
            _cls=_cls,
            service=service,
            instance=instance,
            watcher=watcher,
            bounce_by=bounce_by,
            wait_until=wait_until,
            failures=failures,
            bounce_timers=bounce_timers,
            processed_count=processed_count,
        )


class PaastaThread(Thread):
    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)


class PaastaQueue(Queue):
    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        self.name = name
        super().__init__(*args, **kwargs)

    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)

    def put(self, item: Any, *args: Any, **kwargs: Any) -> None:
        self.log.debug(f"Adding {item} to {self.name} queue")
        super().put(item, *args, **kwargs)


def exponential_back_off(
    failures: int, factor: float, base: float, max_time: float
) -> float:
    seconds = factor * base ** failures
    return seconds if seconds < max_time else max_time


def get_service_instances_needing_update(
    marathon_clients: MarathonClients,
    instances: Collection[Tuple[str, str]],
    cluster: str,
) -> List[Tuple[str, str, MarathonServiceConfig]]:
    marathon_apps = {}
    for marathon_client in marathon_clients.get_all_clients():
        marathon_apps.update(
            {app.id: app for app in get_all_marathon_apps(marathon_client)}
        )

    marathon_app_ids = marathon_apps.keys()
    service_instances = []
    for service, instance in instances:
        try:
            config = load_marathon_service_config_no_cache(
                service=service,
                instance=instance,
                cluster=cluster,
                soa_dir=DEFAULT_SOA_DIR,
            )
            config_app = config.format_marathon_app_dict()
            app_id = "/{}".format(config_app["id"])
        # Not ideal but we rely on a lot of user input to create the app dict
        # and we really can't afford to bail if just one app definition is malformed
        except Exception as e:
            print(
                "ERROR: Skipping {}.{} because: '{}'".format(service, instance, str(e))
            )
            continue
        if app_id not in marathon_app_ids:
            service_instances.append((service, instance, config))
        elif marathon_apps[app_id].instances != config_app["instances"]:
            service_instances.append((service, instance, config))
    return service_instances


def get_marathon_clients_from_config() -> MarathonClients:
    system_paasta_config = load_system_paasta_config()
    marathon_servers = get_marathon_servers(system_paasta_config)
    marathon_clients = get_marathon_clients(marathon_servers)
    return marathon_clients


class DelayDeadlineQueue:
    """Entries into this queue have both a wait_until and a bounce_by. Before wait_until, get() will not return an entry.
    get() returns the entry whose wait_until has passed and which has the lowest bounce_by."""

    def __init__(self) -> None:
        self.available_service_instances: PriorityQueue[
            Tuple[float, ServiceInstance]
        ] = PriorityQueue()
        self.unavailable_service_instances: PriorityQueue[
            Tuple[float, float, ServiceInstance]
        ] = PriorityQueue()

        self.unavailable_service_instances_modify = Condition()
        self.background_thread_started = Event()
        Thread(target=self.move_from_unavailable_to_available, daemon=True).start()
        self.background_thread_started.wait()

    @property
    def log(self) -> logging.Logger:
        name = ".".join([type(self).__module__, type(self).__name__])
        return logging.getLogger(name)

    def put(self, si: ServiceInstance) -> None:
        self.log.debug(
            f"adding {si.service}.{si.instance} to queue with wait_until {si.wait_until} and bounce_by {si.bounce_by}"
        )
        with self.unavailable_service_instances_modify:
            self.unavailable_service_instances.put((si.wait_until, si.bounce_by, si))
            self.unavailable_service_instances_modify.notify()

    def move_from_unavailable_to_available(self) -> None:
        self.background_thread_started.set()
        with self.unavailable_service_instances_modify:
            while True:
                try:
                    while True:
                        wait_until, bounce_by, si = (
                            self.unavailable_service_instances.get_nowait()
                        )
                        if wait_until < time.time():
                            self.available_service_instances.put_nowait((bounce_by, si))
                        else:
                            self.unavailable_service_instances.put_nowait(
                                (wait_until, bounce_by, si)
                            )
                            timeout = wait_until - time.time()
                            break
                except Empty:
                    timeout = None

                self.unavailable_service_instances_modify.wait(timeout=timeout)

    def get(self, block: bool = True, timeout: float = None) -> ServiceInstance:
        _, si = self.available_service_instances.get(block=block, timeout=timeout)
        return si



import collections
import hashlib
import io
import ipaddress
import itertools
import json
import logging
import os.path
import re
from contextlib import contextmanager

from paasta_tools import iptables
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.marathon_tools import get_all_namespaces_for_service
from paasta_tools.utils import get_running_mesos_docker_containers
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import timed_flock


PRIVATE_IP_RANGES = (
    "127.0.0.0/255.0.0.0",
    "10.0.0.0/255.0.0.0",
    "172.16.0.0/255.240.0.0",
    "192.168.0.0/255.255.0.0",
    "169.254.0.0/255.255.0.0",
)
DEFAULT_SYNAPSE_SERVICE_DIR = "/var/run/synapse/services"
DEFAULT_FIREWALL_FLOCK_PATH = "/var/lib/paasta/firewall.flock"
DEFAULT_FIREWALL_FLOCK_TIMEOUT_SECS = 5

RESOLV_CONF = "/etc/resolv.conf"
# not exactly correct, but sufficient to filter out ipv6 or other weird things
IPV4_REGEX = re.compile(r"[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$")


log = logging.getLogger(__name__)


class ServiceGroup(collections.namedtuple("ServiceGroup", ("service", "instance"))):
    """A service group.

    :param service: service name
    :param instance: instance name
    """

    __slots__ = ()

    @property
    def chain_name(self):
        """Return iptables chain name.

        Chain names are limited to 28 characters, so we have to trim quite a
        bit. To attempt to ensure we don't have collisions due to shortening,
        we append a hash to the end.
        """
        chain = "PAASTA.{}".format(self.service[:10])
        chain += "." + hashlib.sha256(json.dumps(self).encode("utf8")).hexdigest()[:10]
        assert len(chain) <= 28, len(chain)
        return chain

    def get_rules(self, soa_dir, synapse_service_dir):
        try:
            conf = get_instance_config(
                self.service,
                self.instance,
                load_system_paasta_config().get_cluster(),
                load_deployments=False,
                soa_dir=soa_dir,
            )
        except NotImplementedError:
            # PAASTA-11414: new instance types may not provide this configuration information;
            # we don't want to break all of the firewall infrastructure when that happens
            return ()
        except NoConfigurationForServiceError:
            # PAASTA-12050: a deleted service may still have containers running on PaaSTA hosts
            # for several minutes after the directory disappears from soa-configs.
            return ()

        if not conf.get_outbound_firewall():
            return ()

        rules = list(_default_rules(conf, self.log_prefix))
        rules.extend(_well_known_rules(conf))
        rules.extend(_smartstack_rules(conf, soa_dir, synapse_service_dir))
        rules.extend(_cidr_rules(conf))
        return tuple(rules)

    def update_rules(self, soa_dir, synapse_service_dir):
        iptables.ensure_chain(
            self.chain_name, self.get_rules(soa_dir, synapse_service_dir)
        )
        iptables.reorder_chain(self.chain_name)

    @property
    def log_prefix(self):
        # log-prefix is limited to 29 characters total
        # space at the end is necessary to separate it from the rest of the line
        # no restrictions on any particular characters afaict
        return f"paasta.{self.service}"[:28] + " "


def _default_rules(conf, log_prefix):
    log_rule = iptables.Rule(
        protocol="ip",
        src="0.0.0.0/0.0.0.0",
        dst="0.0.0.0/0.0.0.0",
        target="LOG",
        target_parameters=(("log-prefix", (log_prefix,)),),
        matches=(("limit", (("limit", ("1/sec",)), ("limit-burst", ("1",)))),),
    )

    policy = conf.get_outbound_firewall()
    if policy == "block":
        return (
            iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst="0.0.0.0/0.0.0.0",
                target="REJECT",
                matches=(),
                target_parameters=((("reject-with", ("icmp-port-unreachable",))),),
            ),
            log_rule,
        )
    elif policy == "monitor":
        return (log_rule,)
    else:
        raise AssertionError(policy)


def _well_known_rules(conf):
    # Allow access to certain resources for all services by default.
    yield iptables.Rule(
        protocol="ip",
        src="0.0.0.0/0.0.0.0",
        dst="0.0.0.0/0.0.0.0",
        target="PAASTA-COMMON",
        matches=(),
        target_parameters=(),
    )

    for dep in conf.get_dependencies() or ():
        resource = dep.get("well-known")
        if resource == "internet":
            yield iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst="0.0.0.0/0.0.0.0",
                target="PAASTA-INTERNET",
                matches=(),
                target_parameters=(),
            )
        elif resource is not None:
            # TODO: handle better
            raise AssertionError(resource)


def _synapse_backends(synapse_service_dir, namespace):
    # Return the contents of the synapse JSON file for a particular service namespace
    # e.g. /var/run/synapse/services/example_happyhour.main.json
    with open(
        os.path.join(synapse_service_dir, namespace + ".json")
    ) as synapse_backend_file:
        synapse_backend_json = json.load(synapse_backend_file)
        return synapse_backend_json


def _yocalhost_rule(port, comment, protocol="tcp"):
    """Return an iptables rule allowing access to a yocalhost port."""
    return iptables.Rule(
        protocol=protocol,
        src="0.0.0.0/0.0.0.0",
        dst="169.254.255.254/255.255.255.255",
        target="ACCEPT",
        matches=(
            ("comment", (("comment", (comment,)),)),
            (protocol, (("dport", (str(port),)),)),
        ),
        target_parameters=(),
    )


def _smartstack_rules(conf, soa_dir, synapse_service_dir):
    for dep in conf.get_dependencies() or ():
        namespace = dep.get("smartstack")
        if namespace is None:
            continue

        # TODO: support wildcards

        # synapse backends
        try:
            backends = _synapse_backends(synapse_service_dir, namespace)
        except (OSError, IOError, ValueError):
            # Don't fatal if something goes wrong loading the synapse files
            log.exception(f"Unable to load backend {namespace}")
            backends = ()

        for backend in backends:
            yield iptables.Rule(
                protocol="tcp",
                src="0.0.0.0/0.0.0.0",
                dst="{}/255.255.255.255".format(backend["host"]),
                target="ACCEPT",
                matches=(
                    ("comment", (("comment", ("backend " + namespace,)),)),
                    ("tcp", (("dport", (str(backend["port"]),)),)),
                ),
                target_parameters=(),
            )

        # synapse-haproxy proxy_port
        service, _ = namespace.split(".", 1)
        service_namespaces = get_all_namespaces_for_service(service, soa_dir=soa_dir)
        port = dict(service_namespaces)[namespace]["proxy_port"]
        yield _yocalhost_rule(port, "proxy_port " + namespace)


def _ports_valid(ports):
    for port in ports:
        try:
            port = int(port)
        except ValueError:
            log.exception(f"Unable to parse port: {port}")
            return False

        if not 1 <= port <= 65535:
            log.error(f"Bogus port number: {port}")
            return False
    else:
        return True


def _cidr_rules(conf):
    for dep in conf.get_dependencies() or ():
        cidr = dep.get("cidr")
        port_str = dep.get("port")

        if cidr is None:
            continue

        try:
            network = ipaddress.IPv4Network(cidr)
        except ipaddress.AddressValueError:
            log.exception(f"Unable to parse IP network: {cidr}")
            continue

        if port_str is not None:
            # port can be either a single port like "443" or a range like "1024:65535"
            ports = str(port_str).split(":")

            if len(ports) > 2:
                log.error(
                    f'"port" must be either a single value or a range like "1024:65535": {port_str}'
                )
                continue

            if not _ports_valid(ports):
                continue

        # Set up an ip rule if no port, or a tcp/udp rule if there is a port
        dst = f"{network.network_address.exploded}/{network.netmask}"
        if port_str is None:
            yield iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst=dst,
                target="ACCEPT",
                matches=(("comment", (("comment", (f"allow {network}:*",)),)),),
                target_parameters=(),
            )
        else:
            for proto in ("tcp", "udp"):
                yield iptables.Rule(
                    protocol=proto,
                    src="0.0.0.0/0.0.0.0",
                    dst=dst,
                    target="ACCEPT",
                    matches=(
                        ("comment", (("comment", (f"allow {network}:{port_str}",)),)),
                        (proto, (("dport", (str(port_str),)),)),
                    ),
                    target_parameters=(),
                )


def services_running_here():
    """Generator helper that yields (service, instance, mac address) of both
    marathon tasks.
    """
    for container in get_running_mesos_docker_containers():
        if container["HostConfig"]["NetworkMode"] != "bridge":
            continue

        service = container["Labels"].get("paasta_service")
        instance = container["Labels"].get("paasta_instance")

        if service is None or instance is None:
            continue

        network_info = container["NetworkSettings"]["Networks"]["bridge"]

        mac = network_info["MacAddress"]
        ip = network_info["IPAddress"]
        yield service, instance, mac, ip


def active_service_groups():
    """Return active service groups."""
    service_groups = collections.defaultdict(set)
    for service, instance, mac, ip in services_running_here():
        # TODO: only include macs that start with MAC_ADDRESS_PREFIX?
        service_groups[ServiceGroup(service, instance)].add(mac)
    return service_groups


def _dns_servers():
    with io.open(RESOLV_CONF) as f:
        for line in f:
            parts = line.split()
            if (
                len(parts) == 2
                and parts[0] == "nameserver"
                and IPV4_REGEX.match(parts[1])
            ):
                yield parts[1]


def ensure_shared_chains():
    _ensure_dns_chain()
    _ensure_internet_chain()
    _ensure_common_chain()


def _ensure_common_chain():
    """The common chain allows access for all services to certain resources."""
    iptables.ensure_chain(
        "PAASTA-COMMON",
        (
            # Allow return traffic for incoming connections
            iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst="0.0.0.0/0.0.0.0",
                target="ACCEPT",
                matches=(("conntrack", (("ctstate", ("ESTABLISHED",)),)),),
                target_parameters=(),
            ),
            _yocalhost_rule(1463, "scribed"),
            _yocalhost_rule(8125, "metrics-relay", protocol="udp"),
            _yocalhost_rule(3030, "sensu"),
            iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst="0.0.0.0/0.0.0.0",
                target="PAASTA-DNS",
                matches=(),
                target_parameters=(),
            ),
        ),
    )


def _ensure_dns_chain():
    iptables.ensure_chain(
        "PAASTA-DNS",
        tuple(
            itertools.chain.from_iterable(
                (
                    iptables.Rule(
                        protocol="udp",
                        src="0.0.0.0/0.0.0.0",
                        dst=f"{dns_server}/255.255.255.255",
                        target="ACCEPT",
                        matches=(("udp", (("dport", ("53",)),)),),
                        target_parameters=(),
                    ),
                    # DNS goes over TCP sometimes, too!
                    iptables.Rule(
                        protocol="tcp",
                        src="0.0.0.0/0.0.0.0",
                        dst=f"{dns_server}/255.255.255.255",
                        target="ACCEPT",
                        matches=(("tcp", (("dport", ("53",)),)),),
                        target_parameters=(),
                    ),
                )
                for dns_server in _dns_servers()
            )
        ),
    )


def _ensure_internet_chain():
    iptables.ensure_chain(
        "PAASTA-INTERNET",
        (
            iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst="0.0.0.0/0.0.0.0",
                target="ACCEPT",
                matches=(),
                target_parameters=(),
            ),
        )
        + tuple(
            iptables.Rule(
                protocol="ip",
                src="0.0.0.0/0.0.0.0",
                dst=ip_range,
                target="RETURN",
                matches=(),
                target_parameters=(),
            )
            for ip_range in PRIVATE_IP_RANGES
        ),
    )


def ensure_service_chains(service_groups, soa_dir, synapse_service_dir):
    """Ensure service chains exist and have the right rules.

    service_groups is a dict {ServiceGroup: set([mac_address..])}

    Returns dictionary {[service chain] => [list of mac addresses]}.
    """
    chains = {}
    for service, macs in service_groups.items():
        service.update_rules(soa_dir, synapse_service_dir)
        chains[service.chain_name] = macs
    return chains


def dispatch_rule(chain, mac):
    return iptables.Rule(
        protocol="ip",
        src="0.0.0.0/0.0.0.0",
        dst="0.0.0.0/0.0.0.0",
        target=chain,
        matches=(("mac", (("mac-source", (mac.upper(),)),)),),
        target_parameters=(),
    )


def ensure_dispatch_chains(service_chains):
    paasta_rules = set(
        itertools.chain.from_iterable(
            (dispatch_rule(chain, mac) for mac in macs)
            for chain, macs in service_chains.items()
        )
    )
    iptables.ensure_chain("PAASTA", paasta_rules)

    jump_to_paasta = iptables.Rule(
        protocol="ip",
        src="0.0.0.0/0.0.0.0",
        dst="0.0.0.0/0.0.0.0",
        target="PAASTA",
        matches=(),
        target_parameters=(),
    )
    iptables.ensure_rule("INPUT", jump_to_paasta)
    iptables.ensure_rule("FORWARD", jump_to_paasta)


def garbage_collect_old_service_chains(desired_chains):
    current_paasta_chains = {
        chain for chain in iptables.all_chains() if chain.startswith("PAASTA.")
    }
    for chain in current_paasta_chains - set(desired_chains):
        iptables.delete_chain(chain)


def general_update(soa_dir, synapse_service_dir):
    """Update iptables to match the current PaaSTA state."""
    ensure_shared_chains()
    service_chains = ensure_service_chains(
        active_service_groups(), soa_dir, synapse_service_dir
    )
    ensure_dispatch_chains(service_chains)
    garbage_collect_old_service_chains(service_chains)


def prepare_new_container(soa_dir, synapse_service_dir, service, instance, mac):
    """Update iptables to include rules for a new (not yet running) MAC address
    """
    ensure_shared_chains()  # probably already set, but just to be safe
    service_group = ServiceGroup(service, instance)
    service_group.update_rules(soa_dir, synapse_service_dir)
    iptables.insert_rule("PAASTA", dispatch_rule(service_group.chain_name, mac))


@contextmanager
def firewall_flock(flock_path=DEFAULT_FIREWALL_FLOCK_PATH):
    """ Grab an exclusive flock to avoid concurrent iptables updates
    """
    with io.FileIO(flock_path, "w") as f:
        with timed_flock(f, seconds=DEFAULT_FIREWALL_FLOCK_TIMEOUT_SECS):
            yield



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_tron_namespace.py service [service...] | --all

Deploy a namespace to the local Tron master from a service configuration file.
Reads from the soa_dir /nail/etc/services by default.

The script will load the service configuration file, generate a Tron configuration
file for it, and send the updated file to Tron.
"""
import argparse
import logging
import sys

from paasta_tools import tron_tools
from paasta_tools.tron_tools import MASTER_NAMESPACE

log = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Update the Tron namespace configuration for a service."
    )
    parser.add_argument("services", nargs="*", help="Services to update.")
    parser.add_argument(
        "-a",
        "--all",
        dest="all_namespaces",
        action="store_true",
        help="Update all available Tron namespaces.",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=tron_tools.DEFAULT_SOA_DIR,
        help="Use a different soa config directory",
    )
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument("--dry-run", action="store_true", default=False)
    parser.add_argument(
        "--cluster",
        help="Cluster to read configs for. Defaults to the configuration in /etc/paasta",
        default=None,
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=log_level)

    if not args.cluster:
        args.cluster = tron_tools.load_tron_config().get_cluster_name()

    if args.all_namespaces:
        if args.services:
            log.error("Do not pass service names with --all flag")
            sys.exit(1)

        try:
            services = tron_tools.get_tron_namespaces(
                cluster=args.cluster, soa_dir=args.soa_dir
            )
        except Exception as e:
            log.error("Failed to list tron namespaces: {error}".format(error=str(e)))
            sys.exit(1)
    else:
        services = args.services

    if not services:
        log.warning("No namespaces found")
        sys.exit(0)

    if not args.dry_run:
        client = tron_tools.get_tron_client()

    updated = []
    failed = []
    skipped = []

    master_config = tron_tools.create_complete_master_config(
        cluster=args.cluster, soa_dir=args.soa_dir
    )
    if args.dry_run:
        log.info(f"Would update {MASTER_NAMESPACE} to:")
        log.info(f"{master_config}")
        updated.append(MASTER_NAMESPACE)
    else:
        if client.update_namespace(MASTER_NAMESPACE, master_config):
            updated.append(MASTER_NAMESPACE)
            log.debug(f"Updated {MASTER_NAMESPACE}")
        else:
            skipped.append(MASTER_NAMESPACE)
            log.debug(f"Skipped {MASTER_NAMESPACE}")

    for service in sorted(services):
        try:
            new_config = tron_tools.create_complete_config(
                cluster=args.cluster, service=service, soa_dir=args.soa_dir
            )
            if args.dry_run:
                log.info(f"Would update {service} to:")
                log.info(f"{new_config}")
                updated.append(service)
            else:
                if client.update_namespace(service, new_config):
                    updated.append(service)
                    log.debug(f"Updated {service}")
                else:
                    skipped.append(service)
                    log.debug(f"Skipped {service}")
        except Exception as e:
            log.error(f"Update for {service} failed: {str(e)}")
            log.debug(f"Exception while updating {service}", exc_info=1)
            failed.append(service)

    skipped_report = skipped if args.verbose else len(skipped)
    log.info(
        f"Updated following namespaces: {updated}, "
        f"failed: {failed}, skipped: {skipped_report}"
    )

    sys.exit(1 if failed else 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./paasta_dump_locally_running_services.py [options]

Outputs a JSON-encoded list of services that are running on this host along
with the host port that each service is listening on.

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
"""
import argparse
import json
import sys

from paasta_tools.marathon_tools import get_marathon_services_running_here_for_nerve
from paasta_tools.marathon_tools import get_puppet_services_running_here_for_nerve
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import paasta_print


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description="Dumps information about locally running services."
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    return parser.parse_args(argv)


def main(argv=None):
    args = parse_args(argv)
    soa_dir = args.soa_dir

    service_dump = get_marathon_services_running_here_for_nerve(
        cluster=None, soa_dir=soa_dir
    ) + get_puppet_services_running_here_for_nerve(soa_dir=soa_dir)

    paasta_print(json.dumps(service_dump))
    sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging

from paasta_tools.autoscaling.autoscaling_service_lib import autoscale_services
from paasta_tools.marathon_tools import DEFAULT_SOA_DIR


def parse_args():
    parser = argparse.ArgumentParser(description="Autoscales marathon jobs")
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Increase logging verboseness"
    )
    parser.add_argument(
        "services",
        type=str,
        nargs="*",
        help="name of services to scale (optional defaults to all autoscaling enabled services)",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)
    autoscale_services(soa_dir=args.soa_dir, services=args.services)


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the meat of the logic for most of the scripts
that interact with marathon. There's config parsers, url composers,
and a number of other things used by other components in order to
make the PaaSTA stack work.
"""
import copy
import datetime
import json
import logging
import os
import socket
import sys
from collections import defaultdict
from math import ceil
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeVar

import pytz
import requests
import service_configuration_lib
from marathon import MarathonClient
from marathon import MarathonHttpError
from marathon import NotFoundError
from marathon.models.app import MarathonApp
from marathon.models.app import MarathonTask
from marathon.models.queue import MarathonQueueItem
from mypy_extensions import TypedDict

from paasta_tools.long_running_service_tools import BounceMethodConfigDict
from paasta_tools.long_running_service_tools import InvalidHealthcheckMode
from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.long_running_service_tools import LongRunningServiceConfig
from paasta_tools.long_running_service_tools import LongRunningServiceConfigDict
from paasta_tools.long_running_service_tools import ServiceNamespaceConfig
from paasta_tools.mesos.exceptions import NoSlavesAvailableError
from paasta_tools.mesos_tools import filter_mesos_slaves_by_blacklist
from paasta_tools.mesos_tools import get_mesos_network_for_net
from paasta_tools.mesos_tools import get_mesos_slaves_grouped_by_attribute
from paasta_tools.mesos_tools import mesos_services_running_here
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.secret_tools import get_secret_hashes
from paasta_tools.utils import _log
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import Constraint
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DockerParameter
from paasta_tools.utils import DockerVolume
from paasta_tools.utils import get_code_sha_from_dockerurl
from paasta_tools.utils import get_config_hash
from paasta_tools.utils import get_user_agent
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import MarathonConfigDict
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaNotConfiguredError
from paasta_tools.utils import SystemPaastaConfig
from paasta_tools.utils import time_cache

# Marathon creates Mesos tasks with an id composed of the app's full name, a
# spacer, and a UUID. This variable is that spacer. Note that we don't control
# this spacer, i.e. you can't change it here and expect the world to change
# with you. We need to know what it is so we can decompose Mesos task ids.
MESOS_TASK_SPACER = "."
PUPPET_SERVICE_DIR = "/etc/nerve/puppet_services.d"


# A set of config attributes that don't get included in the hash of the config.
# These should be things that PaaSTA/Marathon knows how to change without requiring a bounce.
CONFIG_HASH_BLACKLIST = {
    "instances",
    "backoff_seconds",
    "min_instances",
    "max_instances",
}

log = logging.getLogger(__name__)
logging.getLogger("marathon").setLevel(logging.WARNING)


class MarathonServers(NamedTuple):
    current: Sequence["MarathonConfig"]
    previous: Sequence["MarathonConfig"]


_RendezvousHashT = TypeVar("_RendezvousHashT")


def rendezvous_hash(
    choices: Sequence[_RendezvousHashT],
    key: str,
    salt: str = "",
    hash_func: Callable[[str], str] = get_config_hash,
) -> _RendezvousHashT:
    """For each choice, calculate the hash of the index of that choice combined with the key, then return the choice
    whose corresponding hash is highest.

    :param choices: A sequence of arbitrary values. The "winning" value will be returned."""
    max_hash_value = None
    max_hash_choice = None

    if len(choices) == 0:
        raise ValueError("Must pass at least one choice to rendezvous_hash")

    for i, choice in enumerate(choices):
        str_to_hash = MESOS_TASK_SPACER.join([str(i), key, salt])
        hash_value = hash_func(str_to_hash)
        if max_hash_value is None or hash_value > max_hash_value:
            max_hash_value = hash_value
            max_hash_choice = choice

    return max_hash_choice


class MarathonClients:
    def __init__(
        self, current: Sequence[MarathonClient], previous: Sequence[MarathonClient]
    ) -> None:
        self.current = current
        self.previous = previous

    def __repr__(self) -> str:
        return f"MarathonClients(current={self.current!r}, previous={self.previous!r}"

    def get_current_client_for_service(
        self, job_config: "MarathonServiceConfig"
    ) -> MarathonClient:
        service_instance = compose_job_id(job_config.service, job_config.instance)
        if job_config.get_marathon_shard() is not None:
            return self.current[job_config.get_marathon_shard()]
        else:
            return rendezvous_hash(choices=self.current, key=service_instance)

    def get_previous_clients_for_service(
        self, job_config: "MarathonServiceConfig"
    ) -> Sequence[MarathonClient]:
        service_instance = compose_job_id(job_config.service, job_config.instance)
        if job_config.get_previous_marathon_shards() is not None:
            return [self.previous[i] for i in job_config.get_previous_marathon_shards()]
        else:
            try:
                return [rendezvous_hash(choices=self.previous, key=service_instance)]
            except ValueError:
                return []

    def get_all_clients_for_service(
        self, job_config: "MarathonServiceConfig"
    ) -> Sequence[MarathonClient]:
        """Return the set of all clients that a service might have apps on, with no duplicate clients."""
        all_clients = [self.get_current_client_for_service(job_config)]
        all_clients.extend(self.get_previous_clients_for_service(job_config))

        return dedupe_clients(all_clients)

    def get_all_clients(self) -> Sequence[MarathonClient]:
        """Return the set of all unique clients."""
        return dedupe_clients(self.current + self.previous)  # type: ignore


def dedupe_clients(all_clients: Iterable[MarathonClient],) -> Sequence[MarathonClient]:
    """Return a subset of the clients with no servers in common. The assumption here is that if there's any overlap in
    servers, then two clients are talking about the same cluster."""
    all_seen_servers: Set[str] = set()
    deduped_clients: List[MarathonClient] = []
    for client in all_clients:
        if not any(server in all_seen_servers for server in client.servers):
            all_seen_servers.update(client.servers)
            deduped_clients.append(client)

    return deduped_clients


def get_marathon_servers(system_paasta_config: SystemPaastaConfig) -> MarathonServers:
    """
    :param system_paasta_config: A SystemPaastaConfig object representing the system
                                 configuration.
    """
    current = [MarathonConfig(x) for x in system_paasta_config.get_marathon_servers()]
    previous = [
        MarathonConfig(x) for x in system_paasta_config.get_previous_marathon_servers()
    ]
    return MarathonServers(current=current, previous=previous)


class MarathonNotConfigured(Exception):
    pass


class AutoscalingParamsDict(TypedDict, total=False):
    metrics_provider: str
    decision_policy: str
    setpoint: float


class MarathonServiceConfigDict(LongRunningServiceConfigDict, total=False):
    autoscaling: AutoscalingParamsDict
    backoff_factor: float
    max_launch_delay_seconds: float
    bounce_method: str
    bounce_health_params: Dict[str, Any]
    bounce_margin_factor: float
    accepted_resource_roles: Optional[List[str]]
    host_port: int
    marathon_shard: int
    previous_marathon_shards: List[int]


class CommandDict(TypedDict):
    value: str


class HealthcheckDict(TypedDict, total=False):
    protocol: str
    gracePeriodSeconds: float
    intervalSeconds: float
    portIndex: int
    timeoutSeconds: float
    maxConsecutiveFailures: int
    path: str
    command: CommandDict


# These are more-or-less duplicated from native_service_config, but this code uses camelCase, not snake_case.
# We could probably refactor this, and just use snake_case -- Mesos is picky, but marathon-python will auto-convert.
class MarathonDockerPortMapping(TypedDict):
    hostPort: int
    containerPort: int
    protocol: str


class MarathonPortDefinition(TypedDict):
    port: int
    protocol: str


class MarathonDockerInfo(TypedDict, total=False):
    image: str
    network: str
    portMappings: List[MarathonDockerPortMapping]
    parameters: Sequence[DockerParameter]


class MarathonContainerInfo(TypedDict):
    type: str
    docker: MarathonDockerInfo
    volumes: List[DockerVolume]


class FormattedMarathonAppDict(BounceMethodConfigDict, total=False):
    container: MarathonContainerInfo
    uris: List[str]
    backoff_seconds: float
    backoff_factor: float
    max_launch_delay_seconds: float
    health_checks: List[HealthcheckDict]
    env: Dict[str, str]
    mem: float
    cpus: float
    gpus: int
    disk: float
    constraints: List[Constraint]
    cmd: str
    args: List[str]
    id: str
    port_definitions: List[MarathonPortDefinition]
    require_ports: bool
    accepted_resource_roles: List[str]


class MarathonConfig(dict):
    def __init__(self, config: MarathonConfigDict) -> None:
        super().__init__(config)

    @property
    def url(self) -> List[str]:
        return self.get_url()

    @property
    def user(self) -> str:
        return self.get_username()

    @property
    def passwd(self) -> str:
        return self.get_password()

    def get_url(self) -> List[str]:
        """Get the Marathon API url

        :returns: The Marathon API endpoint"""
        try:
            return self["url"]
        except KeyError:
            raise MarathonNotConfigured(
                "Could not find marathon url in system marathon config"
            )

    def get_username(self) -> str:
        """Get the Marathon API username

        :returns: The Marathon API username"""
        try:
            return self["user"]
        except KeyError:
            raise MarathonNotConfigured(
                "Could not find marathon user in system marathon config"
            )

    def get_password(self) -> str:
        """Get the Marathon API password

        :returns: The Marathon API password"""
        try:
            return self["password"]
        except KeyError:
            raise MarathonNotConfigured(
                "Could not find marathon password in system marathon config"
            )


def load_marathon_service_config_no_cache(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> "MarathonServiceConfig":
    """Read a service instance's configuration for marathon.

    If a branch isn't specified for a config, the 'branch' key defaults to
    paasta-${cluster}.${instance}.

    :param name: The service name
    :param instance: The instance of the service to retrieve
    :param cluster: The cluster to read the configuration for
    :param load_deployments: A boolean indicating if the corresponding deployments.json for this service
                             should also be loaded
    :param soa_dir: The SOA configuration directory to read from
    :returns: A dictionary of whatever was in the config for the service instance"""
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    marathon_conf_file = "marathon-%s" % cluster
    instance_configs = service_configuration_lib.read_extra_service_information(
        service, marathon_conf_file, soa_dir=soa_dir
    )

    if instance.startswith("_"):
        raise InvalidJobNameError(
            f"Unable to load marathon job config for {service}.{instance} as instance name starts with '_'"
        )
    if instance not in instance_configs:
        raise NoConfigurationForServiceError(
            f"{instance} not found in config file {soa_dir}/{service}/{marathon_conf_file}.yaml."
        )

    general_config = deep_merge_dictionaries(
        overrides=instance_configs[instance], defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = MarathonServiceConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    return MarathonServiceConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )


@time_cache(ttl=5)
def load_marathon_service_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> "MarathonServiceConfig":
    """Read a service instance's configuration for marathon.

    If a branch isn't specified for a config, the 'branch' key defaults to
    paasta-${cluster}.${instance}.

    :param name: The service name
    :param instance: The instance of the service to retrieve
    :param cluster: The cluster to read the configuration for
    :param load_deployments: A boolean indicating if the corresponding deployments.json for this service
                             should also be loaded
    :param soa_dir: The SOA configuration directory to read from
    :returns: A dictionary of whatever was in the config for the service instance"""
    return load_marathon_service_config_no_cache(
        service=service,
        instance=instance,
        cluster=cluster,
        load_deployments=load_deployments,
        soa_dir=soa_dir,
    )


class InvalidMarathonConfig(Exception):
    pass


class MarathonServiceConfig(LongRunningServiceConfig):
    config_dict: MarathonServiceConfigDict

    config_filename_prefix = "marathon"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: MarathonServiceConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )

    def format_cmd(self) -> str:
        cmd = self.get_cmd()
        if cmd is None:
            return None
        if isinstance(cmd, str):
            return cmd
        elif isinstance(cmd, list):
            return " ".join(cmd)
        else:
            raise ValueError("only list or str accepted for cmd")

    def copy(self) -> "MarathonServiceConfig":
        return self.__class__(
            service=self.service,
            instance=self.instance,
            cluster=self.cluster,
            config_dict=self.config_dict.copy(),
            branch_dict=self.branch_dict.copy()
            if self.branch_dict is not None
            else None,
            soa_dir=self.soa_dir,
        )

    def get_autoscaling_params(self) -> AutoscalingParamsDict:
        default_params: AutoscalingParamsDict = {
            "metrics_provider": "mesos_cpu",
            "decision_policy": "proportional",
            "setpoint": 0.8,
        }
        return deep_merge_dictionaries(
            overrides=self.config_dict.get("autoscaling", AutoscalingParamsDict({})),
            defaults=default_params,
        )

    def get_backoff_seconds(self) -> int:
        """backoff_seconds represents a penalization factor for relaunching failing tasks.
        Every time a task fails, Marathon adds this value multiplied by a backoff_factor.
        In PaaSTA we know how many instances a service has, so we adjust the backoff_seconds
        to account for this, which prevents services with large number of instances from
        being penalized more than services with small instance counts. (for example, a service
        with 30 instances will get backed off 10 times faster than a service with 3 instances)."""
        max_instances = self.get_max_instances()
        instances = max_instances if max_instances is not None else self.get_instances()
        if instances == 0:
            return 1
        else:
            return int(ceil(10.0 / instances))

    def get_backoff_factor(self) -> float:
        return self.config_dict.get("backoff_factor", 2)

    def get_max_launch_delay_seconds(self) -> float:
        return self.config_dict.get("max_launch_delay_seconds", 300)

    def get_bounce_method(self) -> str:
        """Get the bounce method specified in the service's marathon configuration.

        :param service_config: The service instance's configuration dictionary
        :returns: The bounce method specified in the config, or 'crossover' if not specified"""
        return self.config_dict.get("bounce_method", "crossover")

    def get_calculated_constraints(
        self,
        system_paasta_config: SystemPaastaConfig,
        service_namespace_config: ServiceNamespaceConfig,
    ) -> List[Constraint]:
        """Gets the calculated constraints for a marathon instance

        If ``constraints`` is specified in the config, it will use that regardless.
        Otherwise it will calculate a good set of constraints from other inputs,
        like ``pool``, blacklist/whitelist, smartstack data, etc.

        :param service_namespace_config: The service instance's configuration dictionary
        :returns: The constraints specified in the config, or defaults described above
        """
        constraints = self.get_constraints()
        if constraints is not None:
            return constraints
        else:
            constraints = self.get_extra_constraints()
            constraints.extend(
                self.get_routing_constraints(
                    service_namespace_config=service_namespace_config,
                    system_paasta_config=system_paasta_config,
                )
            )
            constraints.extend(
                self.get_deploy_constraints(
                    blacklist=self.get_deploy_blacklist(),
                    whitelist=self.get_deploy_whitelist(),
                    system_deploy_blacklist=system_paasta_config.get_deploy_blacklist(),
                    system_deploy_whitelist=system_paasta_config.get_deploy_whitelist(),
                )
            )
            constraints.extend(self.get_pool_constraints())
            constraints.extend(
                self.get_hostname_unique_constraint(
                    system_paasta_config=system_paasta_config
                )
            )
        return constraints

    def get_hostname_unique_constraint(
        self, system_paasta_config: SystemPaastaConfig
    ) -> List[Constraint]:
        """
        "Small" services automatically receive a hostname UNIQUE constraint to reduce
        the risk of all tasks getting launched on the same agent, which might then be lost.

        :param system_paasta_config: A SystemPaastaConfig object representing the system
                                 configuration.
        :returns: a set of constraints for marathon
        """
        auto_hostname_unique_size = system_paasta_config.get_auto_hostname_unique_size()
        app_size = self.get_max_instances() or self.get_desired_instances()
        if app_size <= auto_hostname_unique_size:
            return [["hostname", "UNIQUE"]]
        return []

    def get_routing_constraints(
        self,
        service_namespace_config: ServiceNamespaceConfig,
        system_paasta_config: SystemPaastaConfig,
    ) -> List[Constraint]:
        """
        Returns a set of constraints in order to evenly group a marathon
        application amongst instances of a discovery type.
        If, for example, a given app's 'discover' key is set to 'region', then this function
        computes the constraints required to group the app evenly amongst each
        of the actual 'region' values in the cluster.
        It does so by querying the value of the discover attribute for each expected slave in the cluster (as defined
        by the expected_slave_attributes key in system paasta config), returning a GROUP_BY constraint where the value
        is the number of unique values for that attribute.
        If you have not set expected_slave_attributes in the system paasta config, this function returns an empty list.

        :param service_namespace_config: the config for this service
        :returns: a set of constraints for marathon
        """
        discover_level = service_namespace_config.get_discover()

        expected_slave_attributes = system_paasta_config.get_expected_slave_attributes()
        if expected_slave_attributes is None:
            return []

        fake_slaves = [{"attributes": a} for a in expected_slave_attributes]
        filtered_slaves = filter_mesos_slaves_by_blacklist(
            slaves=fake_slaves,
            blacklist=self.get_deploy_blacklist(),
            whitelist=self.get_deploy_whitelist(),
        )
        # A slave must be allowed by both the instance config's blacklist/whitelist and the system configs' blacklist/
        # whitelist, so we filter twice.
        filtered_slaves = filter_mesos_slaves_by_blacklist(
            slaves=filtered_slaves,
            blacklist=system_paasta_config.get_deploy_blacklist(),
            whitelist=system_paasta_config.get_deploy_whitelist(),
        )

        if not filtered_slaves:
            raise NoSlavesAvailableError(
                (
                    "We do not believe any slaves on the cluster will match the constraints for %s.%s. If you believe "
                    "this is incorrect, have your system administrator adjust the value of expected_slave_attributes "
                    "in the system paasta configs."
                )
                % (self.service, self.instance)
            )

        value_dict = get_mesos_slaves_grouped_by_attribute(
            filtered_slaves, discover_level
        )
        routing_constraints: List[Constraint] = [
            [discover_level, "GROUP_BY", str(len(value_dict.keys()))]
        ]
        return routing_constraints

    def format_marathon_app_dict(self) -> FormattedMarathonAppDict:
        """Create the configuration that will be passed to the Marathon REST API.

        Currently compiles the following keys into one nice dict:

        - id: the ID of the image in Marathon
        - container: a dict containing the docker url and docker launch options. Needed by deimos.
        - uris: blank.
        - ports: an array containing the port.
        - env: environment variables for the container.
        - mem: the amount of memory required.
        - cpus: the number of cpus required.
        - disk: the amount of disk space required.
        - constraints: the constraints on the Marathon app.
        - instances: the number of instances required.
        - cmd: the command to be executed.
        - args: an alternative to cmd that requires the docker container to have an entrypoint.

        The last 7 keys are retrieved using the get_<key> functions defined above.

        :param app_id: The app id
        :param docker_url: The url to the docker image the app will actually execute
        :param docker_volumes: The docker volumes to run the image with, via the
                               marathon configuration file
        :param service_namespace_config: The service instance's configuration dict
        :returns: A dict containing all of the keys listed above"""

        system_paasta_config = load_system_paasta_config()
        docker_url = self.get_docker_url()
        service_namespace_config = load_service_namespace_config(
            service=self.service, namespace=self.get_nerve_namespace()
        )
        docker_volumes = self.get_volumes(
            system_volumes=system_paasta_config.get_volumes()
        )

        net = get_mesos_network_for_net(self.get_net())

        if self.get_container_type() == "DOCKER":
            container_dict: MarathonContainerInfo = {
                "docker": {
                    "image": docker_url,
                    "network": net,
                    "parameters": self.format_docker_parameters(),
                },
                "type": "DOCKER",
                "volumes": docker_volumes,
            }
        else:
            # Only image and forcePullImage are supported: "Mesos Containerizer and Universal Container Runtime"
            # in http://mesosphere.github.io/marathon/1.4/docs/native-docker.html
            container_dict = {  # type: ignore
                "docker": {"image": docker_url},
                "type": "MESOS",
                "volumes": docker_volumes,
            }

        complete_config: FormattedMarathonAppDict = {
            "container": container_dict,
            "uris": [system_paasta_config.get_dockercfg_location()],
            "backoff_seconds": self.get_backoff_seconds(),
            "backoff_factor": self.get_backoff_factor(),
            "max_launch_delay_seconds": self.get_max_launch_delay_seconds(),
            "health_checks": self.get_healthchecks(
                service_namespace_config=service_namespace_config
            ),
            "env": self.get_env(),
            "mem": float(self.get_mem()),
            "cpus": float(self.get_cpus()),
            "disk": float(self.get_disk()),
            "constraints": self.get_calculated_constraints(
                system_paasta_config=system_paasta_config,
                service_namespace_config=service_namespace_config,
            ),
            "instances": self.get_desired_instances(),
            "cmd": self.format_cmd(),
            "args": self.get_args(),
        }

        if self.get_gpus() is not None:
            complete_config["gpus"] = self.get_gpus()

        # Mesos containerizer does not support portMappings
        if net == "BRIDGE" and complete_config["container"]["type"] == "DOCKER":
            complete_config["container"]["docker"]["portMappings"] = [
                {
                    "containerPort": self.get_container_port(),
                    "hostPort": self.get_host_port(),
                    "protocol": "tcp",
                }
            ]
        else:
            complete_config["port_definitions"] = [
                {"port": self.get_host_port(), "protocol": "tcp"}
            ]
            # Without this, we may end up with multiple containers requiring the same port on the same box.
            complete_config["require_ports"] = self.get_host_port() != 0

        accepted_resource_roles = self.get_accepted_resource_roles()
        if accepted_resource_roles is not None:
            complete_config["accepted_resource_roles"] = accepted_resource_roles

        code_sha = get_code_sha_from_dockerurl(docker_url)

        config_hash = get_config_hash(
            self.sanitize_for_config_hash(complete_config, system_paasta_config),
            force_bounce=self.get_force_bounce(),
        )
        complete_config["id"] = format_job_id(
            self.service, self.instance, code_sha, config_hash
        )

        log.debug("Complete configuration for instance is: %s", complete_config)
        return complete_config

    def sanitize_for_config_hash(
        self, config: FormattedMarathonAppDict, system_paasta_config: SystemPaastaConfig
    ) -> Dict[str, Any]:
        """Removes some data from config to make it suitable for
        calculation of config hash.

        Also adds secret HMACs so that we bounce if secret data has changed.
        We need this because the reference to the secret is all Marathon gets
        and this will not change.

        :param config: complete_config hash to sanitize
        :returns: sanitized copy of complete_config hash
        """
        ahash = {
            key: copy.deepcopy(value)
            for key, value in config.items()
            if key not in CONFIG_HASH_BLACKLIST
        }
        ahash["container"]["docker"][  # type: ignore
            "parameters"
        ] = self.format_docker_parameters(with_labels=False)
        secret_hashes = get_secret_hashes(
            environment_variables=config["env"],
            secret_environment=system_paasta_config.get_vault_environment(),
            service=self.service,
            soa_dir=self.soa_dir,
        )
        if secret_hashes:
            ahash["paasta_secrets"] = secret_hashes
        return ahash

    def get_healthchecks(
        self, service_namespace_config: ServiceNamespaceConfig
    ) -> List[HealthcheckDict]:
        """Returns a list of healthchecks per `the Marathon docs`_.

        If you have an http service, it uses the default endpoint that smartstack uses.
        (/status currently)

        Otherwise these do *not* use the same thresholds as smartstack in order to not
        produce a negative feedback loop, where mesos aggressively kills tasks because they
        are slow, which causes other things to be slow, etc.

        If the mode of the service is None, indicating that it was not specified in the service config
        and smartstack is not used by the service, no healthchecks are passed to Marathon. This ensures that
        it falls back to Mesos' knowledge of the task state as described in `the Marathon docs`_.
        In this case, we provide an empty array of healthchecks per `the Marathon API docs`_
        (scroll down to the healthChecks subsection).

        .. _the Marathon docs: https://mesosphere.github.io/marathon/docs/health-checks.html
        .. _the Marathon API docs: https://mesosphere.github.io/marathon/docs/rest-api.html#post-/v2/apps

        :param service_config: service config hash
        :returns: list of healthcheck definitions for marathon"""

        mode = self.get_healthcheck_mode(service_namespace_config)

        graceperiodseconds = self.get_healthcheck_grace_period_seconds()
        intervalseconds = self.get_healthcheck_interval_seconds()
        timeoutseconds = self.get_healthcheck_timeout_seconds()
        maxconsecutivefailures = self.get_healthcheck_max_consecutive_failures()

        if mode == "http" or mode == "https":
            http_path = self.get_healthcheck_uri(service_namespace_config)
            protocol = f"MESOS_{mode.upper()}"
            healthchecks = [
                HealthcheckDict(
                    {
                        "protocol": protocol,
                        "path": http_path,
                        "gracePeriodSeconds": graceperiodseconds,
                        "intervalSeconds": intervalseconds,
                        "portIndex": 0,
                        "timeoutSeconds": timeoutseconds,
                        "maxConsecutiveFailures": maxconsecutivefailures,
                    }
                )
            ]
        elif mode == "tcp":
            healthchecks = [
                HealthcheckDict(
                    {
                        "protocol": "TCP",
                        "gracePeriodSeconds": graceperiodseconds,
                        "intervalSeconds": intervalseconds,
                        "portIndex": 0,
                        "timeoutSeconds": timeoutseconds,
                        "maxConsecutiveFailures": maxconsecutivefailures,
                    }
                )
            ]
        elif mode == "cmd":
            healthchecks = [
                HealthcheckDict(
                    {
                        "protocol": "COMMAND",
                        "command": {"value": self.get_healthcheck_cmd()},
                        "gracePeriodSeconds": graceperiodseconds,
                        "intervalSeconds": intervalseconds,
                        "timeoutSeconds": timeoutseconds,
                        "maxConsecutiveFailures": maxconsecutivefailures,
                    }
                )
            ]
        elif mode is None:
            healthchecks = []
        else:
            raise InvalidHealthcheckMode(
                "Unknown mode: %s. Only acceptable healthcheck modes are http/https/tcp/cmd"
                % mode
            )
        return healthchecks

    def get_bounce_health_params(
        self, service_namespace_config: ServiceNamespaceConfig
    ) -> Dict[str, Any]:
        default: Dict[str, Any] = {}
        if service_namespace_config.is_in_smartstack():
            default = {"check_haproxy": True}
        return self.config_dict.get("bounce_health_params", default)

    def get_bounce_margin_factor(self) -> float:
        return self.config_dict.get("bounce_margin_factor", 1.0)

    def get_accepted_resource_roles(self) -> Optional[List[str]]:
        return self.config_dict.get("accepted_resource_roles", None)

    def get_host_port(self) -> int:
        """Map this port on the host to your container's port 8888. Default is 0, which means Marathon picks a port."""
        return self.config_dict.get("host_port", 0)

    def get_marathon_shard(self) -> Optional[int]:
        """Returns the configured shard of Marathon to use.
        Defaults to None, which means MarathonClients will decide which shard to put this app on."""
        return self.config_dict.get("marathon_shard", None)

    def get_previous_marathon_shards(self) -> Optional[List[int]]:
        """Returns a list of Marathon shards a service might have been on previously.
        Useful for graceful shard migrations. Defaults to None"""
        return self.config_dict.get("previous_marathon_shards", None)


class MarathonDeployStatus:
    """ An enum to represent Marathon app deploy status.
    Changing name of the keys will affect both the paasta CLI and API.
    """

    Running, Deploying, Stopped, Delayed, Waiting, NotRunning = range(0, 6)

    @classmethod
    def tostring(cls, val: int) -> str:
        for k, v in vars(cls).items():
            if v == val:
                return k
        raise ValueError("Unknown Marathon deploy status %d" % val)

    @classmethod
    def fromstring(cls, _str: str) -> int:
        return getattr(cls, _str, None)


def get_marathon_app_deploy_status(
    client: MarathonClient, app: MarathonApp = None
) -> int:
    # Check the launch queue to see if an app is blocked
    is_overdue, backoff_seconds = get_app_queue_status(client, app.id)

    # Based on conditions at https://mesosphere.github.io/marathon/docs/marathon-ui.html
    if is_overdue:
        deploy_status = MarathonDeployStatus.Waiting
    elif backoff_seconds:
        deploy_status = MarathonDeployStatus.Delayed
    elif len(app.deployments) > 0:
        deploy_status = MarathonDeployStatus.Deploying
    elif app.instances == 0 and app.tasks_running == 0:
        deploy_status = MarathonDeployStatus.Stopped
    else:
        deploy_status = MarathonDeployStatus.Running

    return deploy_status


class CachedMarathonClient(MarathonClient):
    @time_cache(ttl=20)
    def list_apps(self, *args: Any, **kwargs: Any) -> Any:
        return super().list_apps(*args, **kwargs)


def get_marathon_client(
    url: Sequence[str], user: str, passwd: str, cached: bool = False
) -> MarathonClient:
    """Get a new Marathon client connection in the form of a MarathonClient object.

    :param url: The url to connect to Marathon at
    :param user: The username to connect with
    :param passwd: The password to connect with
    :param cached: If true, return CachedMarathonClient
    :returns: A new marathon.MarathonClient object"""
    log.info("Connecting to Marathon server at: %s", url)

    session = requests.Session()
    session.headers.update({"User-Agent": get_user_agent()})

    if cached:
        return CachedMarathonClient(url, user, passwd, timeout=30, session=session)
    else:
        return MarathonClient(url, user, passwd, timeout=30, session=session)


def get_marathon_clients(
    marathon_servers: MarathonServers, cached: bool = False
) -> MarathonClients:
    current_servers = marathon_servers.current
    current_clients = []
    for current_server in current_servers:
        current_clients.append(
            get_marathon_client(
                url=current_server.get_url(),
                user=current_server.get_username(),
                passwd=current_server.get_password(),
                cached=cached,
            )
        )
    previous_servers = marathon_servers.previous
    previous_clients = []
    for previous_server in previous_servers:
        previous_clients.append(
            get_marathon_client(
                url=previous_server.get_url(),
                user=previous_server.get_username(),
                passwd=previous_server.get_password(),
                cached=cached,
            )
        )
    return MarathonClients(current=current_clients, previous=previous_clients)


def get_list_of_marathon_clients(
    system_paasta_config: Optional[SystemPaastaConfig] = None, cached: bool = False
) -> Sequence[MarathonClient]:
    if system_paasta_config is None:
        system_paasta_config = load_system_paasta_config()
    marathon_servers = get_marathon_servers(system_paasta_config)
    return get_marathon_clients(marathon_servers, cached=cached).get_all_clients()


def format_job_id(
    service: str,
    instance: str,
    git_hash: Optional[str] = None,
    config_hash: Optional[str] = None,
) -> str:
    """Compose a Marathon app id formatted to meet Marathon's
    `app id requirements <https://mesosphere.github.io/marathon/docs/rest-api.html#id-string>`_

    :param service: The name of the service
    :param instance: The instance of the service
    :param git_hash: The git_hash portion of the job_id. If git_hash is set,
                     config_hash must also be set.
    :param config_hash: The config_hash portion of the job_id. If config_hash
                        is set, git_hash must also be set.
    :returns: a composed app id in a format that Marathon accepts

    """
    service = str(service).replace("_", "--")
    instance = str(instance).replace("_", "--")
    if git_hash:
        git_hash = str(git_hash).replace("_", "--")
    if config_hash:
        config_hash = str(config_hash).replace("_", "--")
    formatted = compose_job_id(service, instance, git_hash, config_hash)
    return formatted


def deformat_job_id(job_id: str) -> Tuple[str, str, str, str]:
    job_id = job_id.replace("--", "_")
    return decompose_job_id(job_id)


def get_all_namespaces_for_service(
    service: str, soa_dir: str = DEFAULT_SOA_DIR, full_name: bool = True
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    """Get all the smartstack namespaces listed for a given service name.

    :param service: The service name
    :param soa_dir: The SOA config directory to read from
    :param full_name: A boolean indicating if the service name should be prepended to the namespace in the
                      returned tuples as described below (Default: True)
    :returns: A list of tuples of the form (service<SPACER>namespace, namespace_config) if full_name is true,
              otherwise of the form (namespace, namespace_config)
    """
    service_config = service_configuration_lib.read_service_configuration(
        service, soa_dir
    )
    smartstack = service_config.get("smartstack", {})
    namespace_list = []
    for namespace in smartstack:
        if full_name:
            name = compose_job_id(service, namespace)
        else:
            name = namespace
        namespace_list.append((name, smartstack[namespace]))
    return namespace_list


def get_all_namespaces(
    soa_dir: str = DEFAULT_SOA_DIR,
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    """Get all the smartstack namespaces across all services.
    This is mostly so synapse can get everything it needs in one call.

    :param soa_dir: The SOA config directory to read from
    :returns: A list of tuples of the form (service.namespace, namespace_config)"""
    rootdir = os.path.abspath(soa_dir)
    namespace_list: List[Tuple[str, ServiceNamespaceConfig]] = []
    for srv_dir in os.listdir(rootdir):
        namespace_list.extend(get_all_namespaces_for_service(srv_dir, soa_dir))
    return namespace_list


def get_app_id_and_task_uuid_from_executor_id(executor_id: str) -> Tuple[str, str]:
    """Parse the marathon executor ID and return the (app id, task uuid)"""
    app_id, task_uuid = executor_id.rsplit(".", 1)
    return app_id, task_uuid


def parse_service_instance_from_executor_id(task_id: str) -> Tuple[str, str]:
    app_id, task_uuid = get_app_id_and_task_uuid_from_executor_id(task_id)
    (srv_name, srv_instance, _, __) = deformat_job_id(app_id)
    return srv_name, srv_instance


def marathon_services_running_here() -> List[Tuple[str, str, int]]:
    """See what marathon services are being run by a mesos-slave on this host.
    :returns: A list of triples of (service, instance, port)"""

    return mesos_services_running_here(
        framework_filter=lambda fw: fw["name"].startswith("marathon"),
        parse_service_instance_from_executor_id=parse_service_instance_from_executor_id,
    )


def get_marathon_services_running_here_for_nerve(
    cluster: Optional[str], soa_dir: str
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    if not cluster:
        try:
            system_paasta_config = load_system_paasta_config()
            cluster = system_paasta_config.get_cluster()
        # In the cases where there is *no* cluster or in the case
        # where there isn't a Paasta configuration file at *all*, then
        # there must be no marathon services running here, so we catch
        # these custom exceptions and return [].
        except (PaastaNotConfiguredError):
            return []
        if not system_paasta_config.get_register_marathon_services():
            return []
    # We try to get the hosts IP here so that we can pass it to Nerve
    # Nerve then sends it as X-Nerve-Check-IP to hacheck. HAProxy also
    # sends a header to hacheck with the backends IP. If we want to cache
    # for haproxy + nerve under the same key (and hence reduce number of
    # total healthchecks) then we need the hosts IP.
    try:
        host_ip = socket.gethostbyname(socket.gethostname())
    except Exception:
        # BUT the show really must go on. if we can't get the hosts IP
        # then hacheck will have to cache two healthchecks per service
        # but atleast we won't fail to register services.
        host_ip = "127.0.0.1"
    # When a cluster is defined in mesos, let's iterate through marathon services
    marathon_services = marathon_services_running_here()
    nerve_list = []
    for name, instance, port in marathon_services:
        try:
            marathon_service_config = load_marathon_service_config(
                service=name,
                instance=instance,
                cluster=cluster,
                load_deployments=False,
                soa_dir=soa_dir,
            )
            for registration in marathon_service_config.get_registrations():
                reg_service, reg_namespace, _, __ = decompose_job_id(registration)
                nerve_dict = load_service_namespace_config(
                    service=reg_service, namespace=reg_namespace, soa_dir=soa_dir
                )
                if not nerve_dict.is_in_smartstack():
                    continue
                nerve_dict["port"] = port
                nerve_dict["paasta_instance"] = instance
                nerve_dict["deploy_group"] = marathon_service_config.get_deploy_group()
                nerve_dict["extra_healthcheck_headers"] = {"X-Nerve-Check-IP": host_ip}
                nerve_list.append((registration, nerve_dict))
        except (KeyError, NoConfigurationForServiceError):
            continue  # SOA configs got deleted for this app, it'll get cleaned up

    return nerve_list


def get_puppet_services_that_run_here() -> Dict[str, List[str]]:
    # find all files in the PUPPET_SERVICE_DIR, but discard broken symlinks
    # this allows us to (de)register services on a machine by
    # breaking/healing a symlink placed by Puppet.
    puppet_service_dir_services = {}
    if os.path.exists(PUPPET_SERVICE_DIR):
        for service_name in os.listdir(PUPPET_SERVICE_DIR):
            if not os.path.exists(os.path.join(PUPPET_SERVICE_DIR, service_name)):
                continue
            with open(os.path.join(PUPPET_SERVICE_DIR, service_name)) as f:
                puppet_service_data = json.load(f)
                puppet_service_dir_services[service_name] = puppet_service_data[
                    "namespaces"
                ]

    return puppet_service_dir_services


def get_puppet_services_running_here_for_nerve(
    soa_dir: str,
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    puppet_services = []
    for service, namespaces in sorted(get_puppet_services_that_run_here().items()):
        for namespace in namespaces:
            puppet_services.append(
                _namespaced_get_classic_service_information_for_nerve(
                    service, namespace, soa_dir
                )
            )
    return puppet_services


def get_classic_service_information_for_nerve(
    name: str, soa_dir: str
) -> Tuple[str, ServiceNamespaceConfig]:
    return _namespaced_get_classic_service_information_for_nerve(name, "main", soa_dir)


def _namespaced_get_classic_service_information_for_nerve(
    name: str, namespace: str, soa_dir: str
) -> Tuple[str, ServiceNamespaceConfig]:
    nerve_dict = load_service_namespace_config(name, namespace, soa_dir)
    port_file = os.path.join(soa_dir, name, "port")
    # If the namespace defines a port, prefer that, otherwise use the
    # service wide port file.
    nerve_dict["port"] = nerve_dict.get(
        "port", None
    ) or service_configuration_lib.read_port(port_file)
    nerve_name = compose_job_id(name, namespace)
    return (nerve_name, nerve_dict)


def get_classic_services_running_here_for_nerve(
    soa_dir: str,
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    classic_services = []
    classic_services_here = service_configuration_lib.services_that_run_here()
    for service in sorted(classic_services_here):
        namespaces = [
            x[0]
            for x in get_all_namespaces_for_service(service, soa_dir, full_name=False)
        ]
        for namespace in namespaces:
            classic_services.append(
                _namespaced_get_classic_service_information_for_nerve(
                    service, namespace, soa_dir
                )
            )
    return classic_services


def list_all_marathon_app_ids(client: MarathonClient,) -> Sequence[str]:
    """List all marathon app_ids, regardless of state

    The raw marathon API returns app ids in their URL form, with leading '/'s
    conforming to the Application Group format:
    https://github.com/mesosphere/marathon/blob/master/docs/docs/application-groups.md

    This function wraps the full output of list_apps to return a list
    in the original form, without leading "/"'s.

    returns: List of app ids in the same format they are POSTed."""
    return [app.id.lstrip("/") for app in get_all_marathon_apps(client)]


def is_app_id_running(app_id: str, client: MarathonClient) -> bool:
    """Returns a boolean indicating if the app is in the current list
    of marathon apps

    :param app_id: The app_id to look for
    :param client: A MarathonClient object"""

    all_app_ids = list_all_marathon_app_ids(client)
    return app_id.lstrip("/") in all_app_ids


def app_has_tasks(
    client: MarathonClient,
    app_id: str,
    expected_tasks: int,
    exact_matches_only: bool = False,
) -> bool:
    """A predicate function indicating whether an app has launched *at least* expected_tasks
    tasks.

    Raises a marathon.NotFoundError when no app with matching id is found.

    :param client: the marathon client
    :param app_id: the app_id to which the tasks should belong. The leading / that marathon appends to
        app_ids is added here.
    :param expected_tasks: the number of tasks to check for
    :param exact_matches_only: a boolean indicating whether we require exactly expected_tasks to be running
    :returns: a boolean indicating whether there are atleast expected_tasks tasks with
        an app id matching app_id
    """
    app_id = "/%s" % app_id
    try:
        tasks = client.list_tasks(app_id=app_id)
    except NotFoundError:
        paasta_print("no app with id %s found" % app_id)
        raise
    paasta_print(
        "app %s has %d of %d expected tasks" % (app_id, len(tasks), expected_tasks)
    )
    if exact_matches_only:
        return len(tasks) == expected_tasks
    else:
        return len(tasks) >= expected_tasks


def get_app_queue(client: MarathonClient, app_id: str) -> Optional[MarathonQueueItem]:
    """Returns the app queue of an application if it exists in Marathon's launch queue

    :param client: The marathon client
    :param app_id: The Marathon app id (without the leading /)
    :returns: The app queue from marathon
    """
    app_id = "/%s" % app_id
    app_queue = client.list_queue(embed_last_unused_offers=True)
    for app_queue_item in app_queue:
        if app_queue_item.app.id == app_id:
            return app_queue_item
    return None


def get_app_queue_status(
    client: MarathonClient, app_id: str
) -> Tuple[Optional[bool], Optional[float]]:
    """Returns the status of an application if it exists in Marathon's launch queue

    :param client: The marathon client
    :param app_id: The Marathon app id (without the leading /)
    :returns: A tuple of the form (is_overdue, current_backoff_delay) or (None, None)
              if the app cannot be found. If is_overdue is True, then Marathon has
              not received a resource offer that satisfies the requirements for the app
    """
    app_queue = get_app_queue(client, app_id)
    return get_app_queue_status_from_queue(app_queue)


def get_app_queue_status_from_queue(
    app_queue_item: Optional[MarathonQueueItem],
) -> Tuple[Optional[bool], Optional[float]]:
    if app_queue_item is None:
        return (None, None)
    return (app_queue_item.delay.overdue, app_queue_item.delay.time_left_seconds)


def get_app_queue_last_unused_offers(
    app_queue_item: Optional[MarathonQueueItem],
) -> Sequence[Dict]:
    """Returns the unused offers for an app

    :param app_queue_item: app_queue_item returned by get_app_queue
    :returns: A list of offers received from mesos, including the reasons they were rejected
    """
    if app_queue_item is None:
        return []
    return app_queue_item.last_unused_offers


def summarize_unused_offers(app_queue: Optional[MarathonQueueItem]) -> Dict[str, int]:
    """Returns a summary of the reasons marathon rejected offers from mesos

    :param app_queue: An app queue item as returned from get_app_queue
    :returns: A dict of rejection_reason: count
    """
    unused_offers = get_app_queue_last_unused_offers(app_queue)
    reasons: Dict[str, int] = defaultdict(lambda: 0)
    for offer in unused_offers:
        for reason in offer["reason"]:
            reasons[reason] += 1
    return reasons


def create_complete_config(
    service: str, instance: str, soa_dir: str = DEFAULT_SOA_DIR
) -> FormattedMarathonAppDict:
    """Generates a complete dictionary to be POST'ed to create an app on Marathon"""
    return load_marathon_service_config(
        service=service,
        instance=instance,
        cluster=load_system_paasta_config().get_cluster(),
        soa_dir=soa_dir,
    ).format_marathon_app_dict()


def get_expected_instance_count_for_namespace(
    service: str, namespace: str, cluster: str = None, soa_dir: str = DEFAULT_SOA_DIR
) -> int:
    """Get the number of expected instances for a namespace, based on the number
    of instances set to run on that namespace as specified in Marathon service
    configuration files.

    :param service: The service's name
    :param namespace: The namespace for that service to check
    :param soa_dir: The SOA configuration directory to read from
    :returns: An integer value of the # of expected instances for the namespace"""
    total_expected = 0
    if not cluster:
        cluster = load_system_paasta_config().get_cluster()

    pscl = PaastaServiceConfigLoader(
        service=service, soa_dir=soa_dir, load_deployments=False
    )
    for job_config in pscl.instance_configs(
        cluster=cluster, instance_type_class=MarathonServiceConfig
    ):
        if f"{service}.{namespace}" in job_config.get_registrations():
            total_expected += job_config.get_instances()
    return total_expected


def get_matching_appids(
    service: str, instance: str, client: MarathonClient, embed_tasks: bool = False
) -> List[str]:
    """Returns a list of appids given a service and instance.
    Useful for fuzzy matching if you think there are marathon
    apps running but you don't know the full instance id"""
    marathon_apps = get_all_marathon_apps(
        client, service_name=service, embed_tasks=embed_tasks
    )
    return [
        app.id for app in marathon_apps if does_app_id_match(service, instance, app.id)
    ]


def get_matching_apps(
    service: str, instance: str, marathon_apps: Sequence[MarathonApp]
) -> Sequence[MarathonApp]:
    """Returns a list of appids given a service and instance.
    Useful for fuzzy matching if you think there are marathon
    apps running but you don't know the full instance id"""
    return [
        app for app in marathon_apps if does_app_id_match(service, instance, app.id)
    ]


def get_matching_apps_with_clients(
    service: str,
    instance: str,
    marathon_apps_with_clients: Sequence[Tuple[MarathonApp, MarathonClient]],
) -> List[Tuple[MarathonApp, MarathonClient]]:
    return [
        (a, c)
        for a, c in marathon_apps_with_clients
        if does_app_id_match(service, instance, a.id)
    ]


def does_app_id_match(service: str, instance: str, app_id: str) -> bool:
    jobid = format_job_id(service, instance)
    expected_prefix = f"/{jobid}{MESOS_TASK_SPACER}"
    return app_id.startswith(expected_prefix)


def get_all_marathon_apps(
    client: MarathonClient,
    service_name: Optional[str] = None,
    embed_tasks: bool = False,
) -> List[MarathonApp]:
    if service_name:
        return client.list_apps(
            embed_tasks=embed_tasks,
            app_id="/" + format_job_id(service=service_name, instance=""),
        )
    else:
        # Ignore apps inside a folder
        return [
            app
            for app in client.list_apps(embed_tasks=embed_tasks)
            if len(app.id.split("/")) <= 2
        ]


def get_marathon_apps_with_clients(
    clients: Sequence[MarathonClient],
    service_name: Optional[str] = None,
    embed_tasks: bool = False,
) -> Sequence[Tuple[MarathonApp, MarathonClient]]:
    marathon_apps_with_clients: List[Tuple[MarathonApp, MarathonClient]] = []
    for client in clients:
        for app in get_all_marathon_apps(client, service_name, embed_tasks=embed_tasks):
            marathon_apps_with_clients.append((app, client))
    return marathon_apps_with_clients


def kill_task(
    client: MarathonClient, app_id: str, task_id: str, scale: bool
) -> Optional[MarathonTask]:
    """Wrapper to the official kill_task method that is tolerant of errors"""
    try:
        return client.kill_task(app_id=app_id, task_id=task_id, scale=True)
    except MarathonHttpError as e:
        # Marathon allows you to kill and scale in one action, but this is not
        # idempotent. If you kill&scale the same task ID twice, the number of instances
        # gets decremented twice. This can lead to a situation where kill&scaling the
        # last task decrements the number of instances below zero, causing an "Object is not
        # valid" message or a "Bean is not valid" message.
        if "is not valid" in e.error_message and e.status_code == 422:
            log.warning(
                "Got 'is not valid' when killing task %s. Continuing anyway." % task_id
            )
            return None
        elif "does not exist" in e.error_message and e.status_code == 404:
            log.warning(
                "Got 'does not exist' when killing task %s. Continuing anyway."
                % task_id
            )
            return None
        else:
            raise


def kill_given_tasks(
    client: MarathonClient, task_ids: Sequence[str], scale: bool
) -> bool:
    """Wrapper to the official kill_given_tasks method that is tolerant of errors"""
    if not task_ids:
        log.debug("No task_ids specified, not killing any tasks")
        return False
    try:
        return client.kill_given_tasks(task_ids=task_ids, scale=scale, force=True)
    except MarathonHttpError as e:
        # Marathon's interface is always async, so it is possible for you to see
        # a task in the interface and kill it, yet by the time it tries to kill
        # it, it is already gone. This is not really a failure condition, so we
        # swallow this error.
        if "is not valid" in e.error_message and e.status_code == 422:
            log.debug("Probably tried to kill a task id that didn't exist. Continuing.")
            return False
        else:
            raise


def is_task_healthy(
    task: MarathonTask, require_all: bool = True, default_healthy: bool = False
) -> bool:
    """Check that a marathon task is healthy

    :param task: the marathon task object
    :param require_all: require all the healthchecks to be passing
        false means that only one needs to pass
    :param default_healthy: cause the function to report healthy if
        there are no health check results
    :returns: True if healthy, False if not"""
    if task.health_check_results:
        results = [hcr.alive for hcr in task.health_check_results]
        if require_all:
            return all(results)
        else:
            return any(results)
    return default_healthy


def is_old_task_missing_healthchecks(task: MarathonTask, app: MarathonApp) -> bool:
    """We check this because versions of Marathon (at least up to 1.1)
    sometimes stop healthchecking tasks, leaving no results. We can normally
    assume that an "old" task which has no healthcheck results is still up
    and healthy but marathon has simply decided to stop healthchecking it.
    """
    health_checks = app.health_checks
    if not task.health_check_results and health_checks and task.started_at:
        if task.started_at.tzinfo:
            started_at_utc = task.started_at.astimezone(pytz.utc)
        else:
            started_at_utc = task.started_at.replace(tzinfo=pytz.utc)

        healthcheck_startup_time = datetime.timedelta(
            seconds=health_checks[0].grace_period_seconds
        ) + datetime.timedelta(seconds=health_checks[0].interval_seconds * 5)
        is_task_old = started_at_utc + healthcheck_startup_time < datetime.datetime.now(
            pytz.utc
        )
        return is_task_old
    return False


def get_num_at_risk_tasks(app: MarathonApp, draining_hosts: Sequence[str]) -> int:
    """Determine how many of an application's tasks are running on
    at-risk (Mesos Maintenance Draining) hosts.

    :param app: A marathon application
    :param draining_hosts: A list of hostnames that are marked as draining.
                           See paasta_tools.mesos_maintenance.get_draining_hosts
    :returns: An integer representing the number of tasks running on at-risk hosts
    """
    hosts_tasks_running_on = [task.host for task in app.tasks]
    num_at_risk_tasks = 0
    for host in hosts_tasks_running_on:
        if host in draining_hosts:
            num_at_risk_tasks += 1
    log.debug("%s has %d tasks running on at-risk hosts." % (app.id, num_at_risk_tasks))
    return num_at_risk_tasks


def broadcast_log_all_services_running_here(
    line: str, component: str = "monitoring"
) -> None:
    """Log a line of text to paasta logs of all services running on this host.

    :param line: text to log
    :param component: paasta log component
    """
    system_paasta_config = load_system_paasta_config()
    cluster = system_paasta_config.get_cluster()

    services = mesos_services_running_here(
        framework_filter=lambda _: True,
        parse_service_instance_from_executor_id=parse_service_instance_from_executor_id,
    )
    for service, instance, _ in services:
        _log(
            line=line,
            service=service,
            instance=instance,
            component=component,
            cluster=cluster,
        )


def broadcast_log_all_services_running_here_from_stdin(
    component: str = "monitoring",
) -> None:
    broadcast_log_all_services_running_here(sys.stdin.read().strip())


def take_up_slack(client: MarathonClient, app: MarathonApp) -> None:
    slack = max(app.instances - len(app.tasks), 0)
    if slack > 0:
        log.info(
            "Scaling %s down from %d to %d instances to remove slack."
            % (app.id, app.instances, app.instances - slack)
        )
        client.scale_app(app_id=app.id, instances=(app.instances - slack), force=True)



#!/usr/bin/env python3
import argparse
import asyncio
import json
import logging
import os.path
import sys
from collections import defaultdict
from enum import Enum
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple

import yaml
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError

logger = logging.getLogger("check_orphans")

PREFIX = "/smartstack/global/"
CHUNK_SIZE = 50  # How many concurrent xinetd connections
DEFAULT_ZK_DISCOVERY_PATH = "/nail/etc/zookeeper_discovery/infrastructure/local.yaml"
DEFAULT_NERVE_XINETD_PORT = 8735


class ExitCode(Enum):
    OK = 0
    ORPHANS = 1
    COLLISIONS = 2


def get_zk_hosts(path: str) -> List[str]:
    with open(path) as f:
        x = yaml.load(f)
    return [f"{host}:{port}" for host, port in x]


SmartstackData = Dict[str, Dict[str, Any]]


def get_zk_data(blacklisted_services: Set[str]) -> SmartstackData:
    logger.info(f"using {DEFAULT_ZK_DISCOVERY_PATH} for zookeeper")
    zk_hosts = get_zk_hosts(DEFAULT_ZK_DISCOVERY_PATH)

    logger.debug(f"connecting to zk hosts {zk_hosts}")
    zk = KazooClient(hosts=zk_hosts)
    zk.start()

    logger.debug(f"pulling smartstack data from zookeeper")
    zk_data = {}
    services = zk.get_children(PREFIX)
    for service in services:
        if service in blacklisted_services:
            continue
        service_instances = zk.get_children(os.path.join(PREFIX, service))
        instances_data = {}
        for instance in service_instances:
            try:
                instance_node = zk.get(os.path.join(PREFIX, service, instance))
            except NoNodeError:
                continue
            instances_data[instance] = json.loads(instance_node[0])
            zk_data[service] = instances_data

    return zk_data


class InstanceTuple(NamedTuple):
    host: str
    port: int
    service: str


def read_from_zk_data(registrations: SmartstackData) -> Set[InstanceTuple]:
    return {
        InstanceTuple(instance_data["host"], instance_data["port"], service)
        for service, instance in registrations.items()
        for instance_data in instance.values()
    }


async def transfer_one_file(
    host: str, port: int = DEFAULT_NERVE_XINETD_PORT
) -> Tuple[str, Optional[str]]:
    logger.debug(f"getting file from {host}")
    try:
        reader, _ = await asyncio.wait_for(
            asyncio.open_connection(host=host, port=port, limit=2 ** 32), timeout=1.0
        )
        resp = await asyncio.wait_for(reader.read(), timeout=1.0)
    except (asyncio.TimeoutError, ConnectionRefusedError):
        logger.warning(f"error getting file from {host}")
        return (host, None)

    return (host, resp.decode())


async def gather_files(hosts: Set[str]) -> Dict[str, str]:
    logger.info("gathering files from {} hosts".format(len(hosts)))
    tasks = [transfer_one_file(host) for host in hosts]
    responses = {}
    for idx in range(0, len(tasks), CHUNK_SIZE):
        resp = await asyncio.gather(
            *tasks[idx : idx + CHUNK_SIZE], return_exceptions=True
        )
        responses.update(dict(resp))
    return responses


def read_one_nerve_file(nerve_config: str) -> Set[InstanceTuple]:
    services = json.loads(nerve_config)["services"]
    return {
        InstanceTuple(
            service["host"], service["port"], service["zk_path"][len(PREFIX) :]
        )
        for service in services.values()
        if service["zk_path"].startswith(PREFIX)
    }


def read_nerve_files(
    nerve_configs: Dict[str, Optional[str]]
) -> Tuple[Set[InstanceTuple], Set[str]]:
    instance_set: Set[InstanceTuple] = set()
    not_found_hosts: Set[str] = set()
    for host, host_config in nerve_configs.items():
        if host_config is None:
            not_found_hosts.add(host)
        else:
            instance_set |= read_one_nerve_file(host_config)
    return instance_set, not_found_hosts


def get_instance_data(
    blacklisted_services: Set[str]
) -> Tuple[Set[InstanceTuple], Set[InstanceTuple]]:
    # Dump ZK
    zk_data = get_zk_data(blacklisted_services)
    zk_instance_data = read_from_zk_data(zk_data)

    hosts = {x[0] for x in zk_instance_data}

    # Dump Nerve configs from each host via xinetd
    results = asyncio.get_event_loop().run_until_complete(gather_files(hosts))

    nerve_instance_data, not_found_hosts = read_nerve_files(results)

    # Filter out anything that we couldn't get a nerve config for
    zk_instance_data_filtered = {
        x for x in zk_instance_data if x[0] not in not_found_hosts
    }

    logger.info("zk_instance_data (unfiltered) len: {}".format(len(zk_instance_data)))
    logger.info(
        "zk_instance_data (filtered) len: {}".format(len(zk_instance_data_filtered))
    )
    logger.info("nerve_instance_data len: {}".format(len(nerve_instance_data)))

    return zk_instance_data_filtered, nerve_instance_data


def check_orphans(
    zk_instance_data: Set[InstanceTuple], nerve_instance_data: Set[InstanceTuple]
) -> ExitCode:
    orphans = zk_instance_data - nerve_instance_data

    # groupby host
    orphans_by_host: DefaultDict[str, List[Tuple[int, str]]] = defaultdict(list)
    for orphan in orphans:
        orphans_by_host[orphan.host].append((orphan.port, orphan.service))

    # collisions
    instance_by_addr: DefaultDict[Tuple[str, int], Set[str]] = defaultdict(set)
    for nerve_inst in nerve_instance_data:
        instance_by_addr[(nerve_inst.host, nerve_inst.port)].add(nerve_inst.service)

    collisions: List[str] = []
    for zk_inst in zk_instance_data:
        nerve_services = instance_by_addr[(zk_inst.host, zk_inst.port)]
        if len(nerve_services) >= 1 and zk_inst.service not in nerve_services:
            collisions.append(
                f"[{zk_inst.host}:{zk_inst.port}] {zk_inst.service} collides with {nerve_services}"
            )

    if collisions:
        logger.warning("Collisions found! Traffic is being misrouted!")
        print("\n".join(collisions))
        return ExitCode.COLLISIONS
    elif orphans:
        logger.warning("{} orphans found".format(len(orphans)))
        print(dict(orphans_by_host))
        return ExitCode.ORPHANS
    else:
        logger.info(
            "No orphans found out of {} service registrations seen".format(
                len(zk_instance_data)
            )
        )
        return ExitCode.OK


def main() -> ExitCode:
    logging.basicConfig(level=logging.WARNING)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--blacklisted-services",
        default="",
        type=str,
        help="Comma separated list of services to ignore",
    )
    args = parser.parse_args()

    zk_instance_data, nerve_instance_data = get_instance_data(
        set(args.blacklisted_services.split(","))
    )

    return check_orphans(zk_instance_data, nerve_instance_data)


if __name__ == "__main__":
    sys.exit(main().value)
